# -*- coding: utf-8 -*-
"""sup3r WindStats module."""
import json
import pandas as pd
import numpy as np
import xarray as xr
import os
import pickle
import logging
import copy
from scipy.ndimage.filters import gaussian_filter
from rex import MultiFileResourceX
from rex.utilities.fun_utils import get_fun_call_str
from sup3r.utilities import ModuleName
from sup3r.utilities.utilities import (get_input_handler_class,
                                       get_source_type,
                                       transform_rotate_wind,
                                       temporal_coarsening,
                                       spatial_coarsening)
from sup3r.qa.utilities import (ws_ramp_rate_dist, tke_series,
                                velocity_gradient_dist, vorticity_dist,
                                tke_wavenumber_spectrum,
                                tke_frequency_spectrum, st_interp)


logger = logging.getLogger(__name__)


class Sup3rWindStats:
    """Class for doing statistical QA on sup3r forward pass wind outputs."""

    def __init__(self, source_file_paths, out_file_path, s_enhance, t_enhance,
                 heights, temporal_slice=slice(None), target=None,
                 shape=None, raster_file=None, qa_fp=None,
                 time_chunk_size=None, cache_pattern=None,
                 overwrite_cache=False, overwrite_stats=False,
                 input_handler=None, output_handler=None, max_workers=None,
                 extract_workers=None, compute_workers=None, load_workers=None,
                 ti_workers=None, get_interp=False, get_hr=True, get_lr=True,
                 include_stats=None, max_values=None, ramp_rate_t_step=1,
                 coarsen=False, smoothing=None, spatial_res=None,
                 max_delta=10):
        """
        Parameters
        ----------
        source_file_paths : list | str
            A list of low-resolution source files to extract raster data from.
            Each file must have the same number of timesteps. Can also pass a
            string with a unix-style file path which will be passed through
            glob.glob
        out_file_path : str
            A single sup3r-resolved output file (either .nc or .h5) with
            high-resolution data corresponding to the
            source_file_paths * s_enhance * t_enhance
        s_enhance : int
            Factor by which the Sup3rGan model will enhance the spatial
            dimensions of low resolution data
        t_enhance : int
            Factor by which the Sup3rGan model will enhance temporal dimension
            of low resolution data
        heights : list
            Heights for which to compute wind stats. e.g. [10, 40, 80, 100]
        temporal_slice : slice | tuple | list
            Slice defining size of full temporal domain. e.g. If we have 5
            files each with 5 time steps then temporal_slice = slice(None) will
            select all 25 time steps. This can also be a tuple / list with
            length 3 that will be interpreted as slice(*temporal_slice)
        target : tuple
            (lat, lon) lower left corner of raster. You should provide
            target+shape or raster_file, or if all three are None the full
            source domain will be used.
        shape : tuple
            (rows, cols) grid size. You should provide target+shape or
            raster_file, or if all three are None the full source domain will
            be used.
        raster_file : str | None
            File for raster_index array for the corresponding target and
            shape. If specified the raster_index will be loaded from the file
            if it exists or written to the file if it does not yet exist.
            If None raster_index will be calculated directly. You should
            provide target+shape or raster_file, or if all three are None the
            full source domain will be used.
        qa_fp : str | None
            Optional filepath to output QA file when you call
            Sup3rWindStats.run()
            (only .pkl is supported)
        time_chunk_size : int
            Size of chunks to split time dimension into for parallel data
            extraction. If running in serial this can be set to the size
            of the full time index for best performance.
        cache_pattern : str | None
            Pattern for files for saving feature data. e.g.
            file_path_{feature}.pkl Each feature will be saved to a file with
            the feature name replaced in cache_pattern. If not None
            feature arrays will be saved here and not stored in self.data until
            load_cached_data is called. The cache_pattern can also include
            {shape}, {target}, {times} which will help ensure unique cache
            files for complex problems.
        overwrite_cache : bool
            Whether to overwrite cache files storing the computed/extracted
            feature data
        overwrite_stats : bool
            Whether to overwrite saved stats
        input_handler : str | None
            data handler class to use for input data. Provide a string name to
            match a class in data_handling.py. If None the correct handler will
            be guessed based on file type and time series properties.
        output_handler : str | None
            data handler class to use for output data. Provide a string name to
            match a class in data_handling.py. If None the correct handler will
            be guessed based on file type and time series properties.
        max_workers : int | None
            Providing a value for max workers will be used to set the value of
            extract_workers, compute_workers, output_workers, and
            load_workers.  If max_workers == 1 then all processes will be
            serialized. If None extract_workers, compute_workers, load_workers,
            output_workers will use their own provided
            values.
        extract_workers : int | None
            max number of workers to use for extracting features from source
            data.
        compute_workers : int | None
            max number of workers to use for computing derived features from
            raw features in source data.
        load_workers : int | None
            max number of workers to use for loading cached feature data.
        ti_workers : int | None
            max number of workers to use to get full time index. Useful when
            there are many input files each with a single time step. If this is
            greater than one, time indices for input files will be extracted in
            parallel and then concatenated to get the full time index. If input
            files do not all have time indices or if there are few input files
            this should be set to one.
        get_interp : bool
            Whether to include interpolated baseline stats in output
        get_hr : bool
            Whether to include high resolution stats in output
        get_lr : bool
            Whether to include low resolution stats in output
        include_stats : list | None
            List of stats to include in output. e.g. ['ramp_rate',
            'velocity_grad', 'vorticity', 'tke_avg_k', 'tke_avg_f', 'tke_ts']
        max_values : dict | None
            Dictionary of max values to keep for stats. e.g.
            {'ramp_rate_max': 10, 'v_grad_max': 14, 'vorticity_max': 7}
        ramp_rate_t_step : int | list
            Number of time steps to use for ramp rate calculation. If low res
            data is hourly then t_step=1 will calculate the hourly ramp rate.
            Can provide a list of different t_step values.
        smoothing : float | None
            Value passed to gaussian filter used for smoothing source data
        coarsen : bool
            Option to coarsen source data according to s_enhance and t_enhance
        spatial_res : float | None
            Spatial resolution for source data in meters. e.g. 2000. This is
            used to determine the wavenumber range for spectra calculations.
        max_delta : int, optional
            Optional maximum limit on the raster shape that is retrieved at
            once. If shape is (20, 20) and max_delta=10, the full raster will
            be retrieved in four chunks of (10, 10). This helps adapt to
            non-regular grids that curve over large distances, by default 20
        """

        logger.info('Initializing Sup3rWindStats and retrieving source data')

        if max_workers is not None:
            extract_workers = compute_workers = load_workers = max_workers
            ti_workers = max_workers

        self.max_values = max_values or {}
        self.ramp_rate_max = self.max_values.get('ramp_rate_max', 10)
        self.v_grad_max = self.max_values.get('v_grad_max', 7)
        self.vorticity_max = self.max_values.get('vorticity_max', 14)
        self.ramp_rate_t_step = (ramp_rate_t_step
                                 if isinstance(ramp_rate_t_step, list)
                                 else [ramp_rate_t_step])
        self.include_stats = include_stats or ['ramp_rate', 'velocity_grad',
                                               'vorticity', 'tke_avg_k',
                                               'tke_avg_f', 'tke_ts',
                                               'mean_ramp_rate']

        self.s_enhance = s_enhance
        self.t_enhance = t_enhance
        self.heights = heights if isinstance(heights, list) else [heights]
        self._out_fp = (out_file_path if isinstance(out_file_path, list)
                        else [out_file_path])
        self._hr_lat_lon = None
        self._hr_time_index = None
        self.get_interp = get_interp
        self.get_hr = get_hr
        self.get_lr = get_lr
        self.cache_pattern = cache_pattern
        self.overwrite_cache = overwrite_cache
        self.overwrite_stats = overwrite_stats
        self.coarsen = coarsen
        self.k_range = None
        self.qa_fp = qa_fp

        source_handler_kwargs = dict(target=target,
                                     shape=shape,
                                     temporal_slice=temporal_slice,
                                     raster_file=raster_file,
                                     cache_pattern=cache_pattern,
                                     time_chunk_size=time_chunk_size,
                                     overwrite_cache=overwrite_cache,
                                     max_workers=max_workers,
                                     extract_workers=extract_workers,
                                     compute_workers=compute_workers,
                                     load_workers=load_workers,
                                     ti_workers=ti_workers,
                                     max_delta=max_delta)
        self.source_data = self.get_source_data(source_file_paths,
                                                input_handler, smoothing,
                                                spatial_res,
                                                source_handler_kwargs)
        if self.get_hr and self._out_fp is not None:
            self.lr_t_slice, self.hr_t_slice = self.time_overlap_slices()
        else:
            self.lr_t_slice = slice(None)
            self.hr_t_slice = slice(None)

        if self.get_hr and self._out_fp is not None:
            if shape is not None or target is not None:
                shape = (self.source_handler.shape[0] * s_enhance,
                         self.source_handler.shape[1] * s_enhance)
                target = self.source_handler.target
            raster_file = (None if raster_file is None
                           else raster_file.replace('.txt', '_hr.txt'))

            output_handler_kwargs = copy.deepcopy(source_handler_kwargs)
            update_kwargs = dict(target=target, shape=shape,
                                 temporal_slice=self.hr_t_slice,
                                 raster_file=raster_file)
            output_handler_kwargs.update(update_kwargs)
            self.output_data = self.get_output_data(self._out_fp,
                                                    output_handler,
                                                    output_handler_kwargs)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

        if type is not None:
            raise

    def close(self):
        """Close any open file handlers"""
        self.output_handler.close()

    def get_source_data(self, source_file_paths, input_handler, smoothing=None,
                        spatial_res=None, source_handler_kwargs=None):
        """Get source data using provided source file paths

        Parameters
        ----------

        source_file_paths : list | str
            A list of low-resolution source files to extract raster data from.
            Each file must have the same number of timesteps. Can also pass a
            string with a unix-style file path which will be passed through
            glob.glob
        input_handler : str | None
            data handler class to use for input data. Provide a string name to
            match a class in data_handling.py. If None the correct handler will
            be guessed based on file type and time series properties.
        smoothing : float | None
            Value passed to gaussian filter used for smoothing source data
        spatial_res : float | None
            Spatial resolution for source data in meters. e.g. 2000. This is
            used to determine the wavenumber range for spectra calculations.
        source_handler_kwargs : dict
            Dictionary of keyword arguments passed to
            `sup3r.preprocessing.data_handling.DataHandler`

        Returns
        -------
        data : ndarray
            Array of data from source file paths
            (spatial_1, spatial_2, temporal, features)
        """

        HandlerClass = get_input_handler_class(source_file_paths,
                                               input_handler)
        self.source_handler = HandlerClass(source_file_paths,
                                           self.features,
                                           val_split=0.0,
                                           **source_handler_kwargs)
        self.source_handler.load_cached_data()
        if self.coarsen:
            logger.info('Coarsening input data with shape='
                        f'{self.source_handler.data.shape}')
            self.source_handler.data = self.coarsen_data(
                self.source_handler.data, smoothing=smoothing)
            logger.info('Coarsened shape='
                        f'{self.source_handler.data.shape}')
        if spatial_res is not None:
            domain_size = spatial_res * self.source_handler.data.shape[1]
            self.k_range = [1 / domain_size, 1 / spatial_res]
        return self.source_handler.data

    def get_output_data(self, out_file_paths, output_handler,
                        output_handler_kwargs=None):
        """Get source data using provided source file paths

        Parameters
        ----------

        out_file_paths : str | list
            A list of sup3r-resolved output files (either .nc or .h5) with
            high-resolution data corresponding to the
            source_file_paths * s_enhance * t_enhance
        output_handler : str | None
            data handler class to use for output data. Provide a string name to
            match a class in data_handling.py. If None the correct handler will
            be guessed based on file type and time series properties.
        output_handler_kwargs : dict
            Dictionary of keyword arguments passed to
            `sup3r.preprocessing.data_handling.DataHandler`

        Returns
        -------
        data : ndarray
            Array of data from output file paths
            (spatial_1, spatial_2, temporal, features)
        """
        shape = output_handler_kwargs.get('shape', None)
        target = output_handler_kwargs.get('target', None)
        if shape is None and target is None:
            self.output_handler = self.output_handler_class(self._out_fp)
            data = np.zeros((self.hr_shape[0], self.hr_shape[1],
                             len(self.hr_time_index[self.hr_t_slice]),
                             len(self.features)), dtype=np.float32)
            for i, height in enumerate(self.heights):
                u, v = self.get_uv_out(height)
                data[..., 2 * i] = u
                data[..., 2 * i + 1] = v
            return data
        else:
            HandlerClass = get_input_handler_class(out_file_paths,
                                                   output_handler)
            self.output_handler = HandlerClass(out_file_paths,
                                               self.features,
                                               val_split=0.0,
                                               **output_handler_kwargs)
            self.output_handler.load_cached_data()
            return self.output_handler.data

    @property
    def hr_shape(self):
        """Shape of output data"""
        shape = (self.lr_shape[0] * self.s_enhance,
                 self.lr_shape[1] * self.s_enhance,
                 self.lr_shape[2] * self.t_enhance)
        return shape

    @property
    def hr_lat_lon(self):
        """Get lat/lon for output data"""
        if self._hr_lat_lon is None:
            meta = self.output_handler.meta
            lats = meta.latitude.values
            lons = meta.longitude.values
            lat_lon = np.dstack((lats.reshape(self.hr_shape[:-1]),
                                 lons.reshape(self.hr_shape[:-1])))
        return lat_lon

    @property
    def meta(self):
        """Get the meta data corresponding to the flattened source low-res data

        Returns
        -------
        pd.DataFrame
        """
        lat_lon = self.source_handler.lat_lon
        meta = pd.DataFrame({'latitude': lat_lon[..., 0].flatten(),
                             'longitude': lat_lon[..., 1].flatten()})
        return meta

    @property
    def lr_shape(self):
        """Get the shape of the source low-res data raster
        (rows, cols, time, features)"""
        return self.source_handler.data[..., self.lr_t_slice, 0].shape

    @property
    def lr_time_index(self):
        """Get the time index associated with the source low-res data

        Returns
        -------
        pd.DatetimeIndex
        """
        return self.source_handler.time_index.tz_localize(None)

    @property
    def hr_time_index(self):
        """Get the time index associated with the high-res data

        Returns
        -------
        pd.DatetimeIndex
        """
        if self._hr_time_index is None:
            if self.output_type == 'nc':
                raise NotImplementedError('Netcdf output not yet supported')
            else:
                ti = self.output_handler_class(self._out_fp).time_index
                self._hr_time_index = ti.tz_localize(None)
        return self._hr_time_index

    def time_overlap_slices(self):
        """Get slices for temporal overlap of low and high resolution data

        Returns
        -------
        lr_slice : slice
            Slice for overlap of low resolution data with high resolution data
        hr_slice : slice
            Slice for overlap of high resolution data with low resolution data
        """
        min_time = np.max((self.lr_time_index[0], self.hr_time_index[0]))
        max_time = np.min((self.lr_time_index[-1], self.hr_time_index[-1]))

        lr_start = next(i for i, t in enumerate(self.lr_time_index)
                        if t >= min_time)
        hr_start = next(i for i, t in enumerate(self.hr_time_index)
                        if t >= min_time)

        lr_end = next(len(self.lr_time_index) - i
                      for i, t in enumerate(self.lr_time_index[::-1])
                      if t <= max_time) + 1
        hr_end = next(len(self.hr_time_index) - i
                      for i, t in enumerate(self.hr_time_index[::-1])
                      if t <= max_time) + 1

        return slice(lr_start, lr_end), slice(hr_start, hr_end)

    @property
    def features(self):
        """Get a list of requested wind feature names

        Returns
        -------
        list
        """

        # all lower case
        features = []
        for height in self.heights:
            features.append(f'U_{height}m')
            features.append(f'V_{height}m')

        return features

    def feature_indices(self, height):
        """Indices for U/V of given height

        Parameters
        ----------
        height : int
            Height in meters for requested U/V fields

        Returns
        -------
        uidx : int
            Index for U_{height}m
        vidx : int
            Index for V_{height}m
        """
        uidx = self.features.index(f'U_{height}m')
        vidx = self.features.index(f'V_{height}m')
        return uidx, vidx

    @property
    def output_type(self):
        """Get output data type

        Returns
        -------
        output_type
            e.g. 'nc' or 'h5'
        """
        ftype = get_source_type(self._out_fp)
        if ftype not in ('nc', 'h5'):
            msg = 'Did not recognize output file type: {}'.format(self._out_fp)
            logger.error(msg)
            raise TypeError(msg)
        return ftype

    @property
    def output_handler_class(self):
        """Get the output handler class.

        Returns
        -------
        HandlerClass : rex.Resource | xr.open_dataset
        """
        if self.output_type == 'nc':
            return xr.open_dataset
        elif self.output_type == 'h5':
            return MultiFileResourceX

    def get_uv_out(self, height):
        """Get an output dataset from the forward pass output file.

        Parameters
        ----------
        name : str
            Name of the output dataset to retrieve. Must be found in the
            features property and the forward pass output file.

        Returns
        -------
        out : np.ndarray
            A copy of the high-resolution output data as a numpy
            array of shape (spatial_1, spatial_2, temporal)
        """

        logger.debug('Getting sup3r u/v data ({}m)'.format(height))
        t_steps = len(self.hr_time_index[self.hr_t_slice])
        shape = f'{self.hr_shape[0]}x{self.hr_shape[1]}x{t_steps}'

        u_file = v_file = None
        if self.cache_pattern is not None:
            tmp_file = self.cache_pattern.replace('{shape}', f'{shape}')
            u_file = tmp_file.replace('{feature}', f'u_{height}m')
            v_file = tmp_file.replace('{feature}', f'v_{height}m')

            if (os.path.exists(u_file) and os.path.exists(v_file)
                    and not self.overwrite_cache):
                u = self.load_cache(u_file)
                v = self.load_cache(v_file)
                return u, v

        if self.output_type == 'nc':
            raise NotImplementedError('Netcdf output not yet supported')
        elif self.output_type == 'h5':
            ws_f = f'windspeed_{height}m'
            wd_f = f'winddirection_{height}m'
            logger.info('Extracting data from output handler for '
                        f'time_slice={self.hr_t_slice}')
            ws = self.output_handler[ws_f, self.hr_t_slice, :]
            wd = self.output_handler[wd_f, self.hr_t_slice, :]
            ws = ws.T.reshape((self.hr_shape[0], self.hr_shape[1], -1))
            wd = wd.T.reshape((self.hr_shape[0], self.hr_shape[1], -1))
            logger.info(f'Transforming ws/wd to u/v for height={height}m '
                        f'with shape={ws.shape}')
            u, v = transform_rotate_wind(ws, wd, self.hr_lat_lon)
            if u_file is not None:
                self.save_cache(u, u_file)
            if v_file is not None:
                self.save_cache(v, v_file)

        return u, v

    @classmethod
    def save_cache(cls, array, file_name):
        """Save data to cache file

        Parameters
        ----------
        array : ndarray
            Wind field data
        file_name : str
            Path to cache file
        """
        logger.info(f'Saving data to {file_name}')
        with open(file_name, 'wb') as f:
            pickle.dump(array, f, protocol=4)

    @classmethod
    def load_cache(cls, file_name):
        """Load data from cache file

        Parameters
        ----------
        file_name : str
            Path to cache file

        Returns
        -------
        array : ndarray
            Wind field data
        """
        logger.info(f'Loading data from {file_name}')
        with open(file_name, 'rb') as f:
            arr = pickle.load(f)
        return arr

    def export(self, qa_fp, data):
        """Export stats dictionary to pkl file.

        Parameters
        ----------
        qa_fp : str | None
            Optional filepath to output QA file (only .h5 is supported)
        data : dict
            A dictionary with stats for low and high resolution wind fields
        """

        if not os.path.exists(qa_fp) or self.overwrite_stats:
            logger.info('Saving windstats output file: "{}"'.format(qa_fp))
            with open(qa_fp, 'wb') as f:
                pickle.dump(data, f)
        else:
            logger.info(f'{qa_fp} already exists. Delete file or run with '
                        'overwrite_stats=True.')

    @classmethod
    def get_node_cmd(cls, config):
        """Get a CLI call to initialize Sup3rWindStats and execute the
        Sup3rWindStats.run() method based on an input config

        Parameters
        ----------
        config : dict
            sup3r wind stats config with all necessary args and kwargs to
            initialize Sup3rWindStats and execute Sup3rWindStats.run()
        """
        import_str = 'import time;\n'
        import_str += 'from reV.pipeline.status import Status;\n'
        import_str += 'from rex import init_logger;\n'
        import_str += 'from sup3r.qa.stats import Sup3rWindStats;\n'

        qa_init_str = get_fun_call_str(cls, config)

        log_file = config.get('log_file', None)
        log_level = config.get('log_level', 'INFO')

        log_arg_str = (f'"sup3r", log_level="{log_level}"')
        if log_file is not None:
            log_arg_str += f', log_file="{log_file}"'

        cmd = (f"python -c \'{import_str}\n"
               "t0 = time.time();\n"
               f"logger = init_logger({log_arg_str});\n"
               f"qa = {qa_init_str};\n"
               "qa.run();\n"
               "t_elap = time.time() - t0;\n")

        job_name = config.get('job_name', None)
        if job_name is not None:
            status_dir = config.get('status_dir', None)
            status_file_arg_str = f'"{status_dir}", '
            status_file_arg_str += f'module="{ModuleName.WIND_STATS}", '
            status_file_arg_str += f'job_name="{job_name}", '
            status_file_arg_str += 'attrs=job_attrs'

            cmd += ('job_attrs = {};\n'.format(json.dumps(config)
                                               .replace("null", "None")
                                               .replace("false", "False")
                                               .replace("true", "True")))
            cmd += 'job_attrs.update({"job_status": "successful"});\n'
            cmd += 'job_attrs.update({"time": t_elap});\n'
            cmd += f'Status.make_job_file({status_file_arg_str})'

        cmd += (";\'\n")

        return cmd.replace('\\', '/')

    def get_spectra_stats(self, u, v, res='low'):
        """Compute spectra based statistics

        Parameters
        ----------
        u: ndarray
            Longitudinal velocity component
            (lat, lon, temporal)
        v : ndarray
            Latitudinal velocity component
            (lat, lon, temporal)

        Returns
        -------
        stats : dict
            Dictionary of spectra stats for wind fields
        """

        k_range = None
        if self.k_range is not None:
            k_range = [self.k_range[0], self.k_range[1]]
            if res == 'high':
                k_range[1] *= self.s_enhance

        stats_dict = {}
        if 'tke_ts' in self.include_stats:
            logger.info('Computing mean kinetic energy spectrum series.')
            stats_dict['tke_ts'] = tke_series(u, v)

        if 'tke_avg_k' in self.include_stats:
            logger.info('Computing time averaged kinetic energy'
                        f' wavenumber spectrum for res={res}. '
                        f'Using k_range={k_range}.')
            stats_dict['tke_avg_k'] = tke_wavenumber_spectrum(
                np.mean(u, axis=-1), np.mean(v, axis=-1), k_range=k_range)

        if 'tke_avg_f' in self.include_stats:
            logger.info('Computing time averaged kinetic energy'
                        ' frequency spectrum.')
            stats_dict['tke_avg_f'] = tke_frequency_spectrum(u, v)

        return stats_dict

    def coarsen_data(self, data, smoothing=None):
        """Re-coarsen a high-resolution synthetic output dataset

        Parameters
        ----------
        data : np.ndarray
            A copy of the high-resolution output data as a numpy
            array of shape (spatial_1, spatial_2, temporal)

        Returns
        -------
        data : np.ndarray
            A spatiotemporally coarsened copy of the input dataset, still with
            shape (spatial_1, spatial_2, temporal)
        """
        n_lats = self.s_enhance * (data.shape[0] // self.s_enhance)
        n_lons = self.s_enhance * (data.shape[1] // self.s_enhance)
        data = spatial_coarsening(data[:n_lats, :n_lons],
                                  s_enhance=self.s_enhance,
                                  obs_axis=False)

        # t_coarse needs shape to be 5D: (obs, s1, s2, t, f)
        data = np.expand_dims(data, axis=0)
        data = temporal_coarsening(data, t_enhance=self.t_enhance)
        data = data[0]

        if smoothing is not None:
            for i in range(data.shape[-1]):
                for t in range(data.shape[-2]):
                    data[..., t, i] = gaussian_filter(data[..., t, i],
                                                      smoothing,
                                                      mode='nearest')

        return data

    def get_ramp_rate_stats(self, u, v, scale=1):
        """Compute statistics for ramp rates

        Parameters
        ----------
        u: ndarray
            Longitudinal velocity component
            (lat, lon, temporal)
        v : ndarray
            Latitudinal velocity component
            (lat, lon, temporal)
        res : str
            Resolution of input fields. If this is 'high' then the ramp rate
            time step needs to be multipled by the temporal enhancement factor

        Returns
        -------
        stats : dict
            Dictionary of ramp rate stats for wind fields
        """

        stats_dict = {}
        if 'ramp_rate' in self.include_stats:
            for i, time in enumerate(self.ramp_rate_t_step):
                logger.info('Computing ramp rate pdf.')
                out = ws_ramp_rate_dist(u, v, diff_max=self.ramp_rate_max,
                                        t_steps=time, scale=scale)
                stats_dict[f'ramp_rate_{self.ramp_rate_t_step[i]}'] = out
        if 'mean_ramp_rate' in self.include_stats:
            for i, time in enumerate(self.ramp_rate_t_step):
                logger.info('Computing mean ramp rate pdf.')
                out = ws_ramp_rate_dist(np.mean(u, axis=(0, 1)),
                                        np.mean(v, axis=(0, 1)),
                                        diff_max=self.ramp_rate_max,
                                        t_steps=time, scale=scale)
                stats_dict[f'mean_ramp_rate_{self.ramp_rate_t_step[i]}'] = out
        return stats_dict

    def get_wind_stats(self, u, v, res='low'):
        """Get stats for wind fields

        Parameters
        ----------
        u: ndarray
            Longitudinal velocity component
            (lat, lon, temporal)
        v : ndarray
            Latitudinal velocity component
            (lat, lon, temporal)
        res : str
            Resolution of input fields. If this is 'high' then the ramp rate
            time step needs to be multipled by the temporal enhancement factor

        Returns
        -------
        stats : dict
            Dictionary of stats for wind fields
        """
        stats_dict = self.get_spectra_stats(u, v, res=res)

        scale = 1 if res == 'high' else self.s_enhance
        if 'velocity_grad' in self.include_stats:
            logger.info('Computing velocity gradient pdf.')
            stats_dict['velocity_grad'] = velocity_gradient_dist(
                u, diff_max=self.v_grad_max, scale=scale)

        if 'mean_velocity_grad' in self.include_stats:
            logger.info('Computing mean velocity gradient pdf.')
            stats_dict['mean_velocity_grad'] = velocity_gradient_dist(
                np.mean(u, axis=-1), diff_max=self.v_grad_max, scale=scale)

        if 'vorticity' in self.include_stats:
            logger.info('Computing vorticity pdf.')
            stats_dict['vorticity'] = vorticity_dist(
                u, v, diff_max=self.vorticity_max, scale=scale)

        if 'mean_vorticity' in self.include_stats:
            logger.info('Computing mean vorticity pdf.')
            stats_dict['mean_vorticity'] = vorticity_dist(
                np.mean(u, axis=-1), np.mean(v, axis=-1),
                diff_max=self.vorticity_max, scale=scale)

        scale = 1 if res == 'high' else self.t_enhance
        out = self.get_ramp_rate_stats(u, v, scale=scale)
        stats_dict.update(out)
        return stats_dict

    def get_height_stats(self, height):
        """Get stats for high and low resolution wind fields

        Parameters
        ----------
        height : int
            Height in meters for requested U/V fields

        Returns
        -------
        low_res : dict
            Dictionary of stats for low resolution wind fields
        high_res : dict
            Dictionary of stats for high resolution wind fields
        interp : dict
            Dictionary of stats for spatiotemporally interpolated wind fields
        """
        low_res = {}
        if self.get_lr:
            uidx, vidx = self.feature_indices(height)
            u_lr = self.source_data[..., self.lr_t_slice, uidx]
            v_lr = self.source_data[..., self.lr_t_slice, vidx]
            logger.info(f'Getting low res stats for height={height}m')
            low_res = self.get_wind_stats(u_lr, v_lr, res='low')

        high_res = {}
        if self.get_hr:
            uidx, vidx = self.feature_indices(height)
            u_hr = self.output_data[..., uidx]
            v_hr = self.output_data[..., vidx]
            logger.info(f'Getting high res stats for height={height}m')
            high_res = self.get_wind_stats(u_hr, v_hr, res='high')

        interp = {}
        if self.get_interp:
            logger.info(f'Interpolating low res U for height={height}')
            u_itp = st_interp(u_lr, self.s_enhance, self.t_enhance)
            logger.info(f'Interpolating low res V for height={height}')
            v_itp = st_interp(v_lr, self.s_enhance, self.t_enhance)
            logger.info('Getting interpolated baseline stats for '
                        f'height={height}m')
            interp = self.get_wind_stats(u_itp, v_itp, res='high')
        return low_res, high_res, interp

    def run(self):
        """Go through all datasets and get the dictionary of wind field
        statistics.

        Returns
        -------
        stats : dict
            Dictionary of statistics, where keys are lr/hr/interp appended with
            the height of the corresponding wind field. Values are dictionaries
            of statistics, such as velocity_gradient, vorticity, ramp_rate,
            etc
        """

        stats = {}
        for idf, height in enumerate(self.heights):
            logger.info('Running WindStats on height {} of {} ({}m)'
                        .format(idf + 1, len(self.heights), height))
            lr_stats, hr_stats, interp = self.get_height_stats(height)

            if self.get_lr:
                stats[f'lr_{height}m'] = lr_stats
            if self.get_hr:
                stats[f'hr_{height}m'] = hr_stats
            if self.get_interp:
                stats[f'interp_{height}m'] = interp

        if self.qa_fp is not None:
            self.export(self.qa_fp, stats)
        logger.info('Finished Sup3rWindStats run method.')

        return stats



# -*- coding: utf-8 -*-
"""Sup3r model software"""
import logging
import tensorflow as tf

from sup3r.models.base import Sup3rGan


logger = logging.getLogger(__name__)


class SolarCC(Sup3rGan):
    """Solar climate change model.

    Modifications to standard Sup3rGan:
        - Content loss is only on the n_days of the center 8 daylight hours of
          the daily true+synthetic high res samples
        - Discriminator only sees n_days of the center 8 daylight hours of the
          daily true high res sample.
        - Discriminator sees random n_days of 8-hour samples of the daily
          synthetic high res sample.
    """

    # starting hour is the hour that daylight starts at, daylight hours is the
    # number of daylight hours to sample, so for example if 8 and 8, the
    # daylight slice will be slice(8, 16). The stride length is the step size
    # for sampling the temporal axis of the generated data to send to the
    # discriminator for the adverserial loss component of the generator. For
    # example, if the generator produces 24 timesteps and stride is 4 and the
    # daylight hours is 8, slices of (0, 8) (4, 12), (8, 16), (12, 20), and
    # (16, 24) will be sent to the disc.
    STARTING_HOUR = 8
    DAYLIGHT_HOURS = 8
    STRIDE_LEN = 4

    @tf.function
    def calc_loss(self, hi_res_true, hi_res_gen, weight_gen_advers=0.001,
                  train_gen=True, train_disc=False):
        """Calculate the GAN loss function using generated and true high
        resolution data.

        Parameters
        ----------
        hi_res_true : tf.Tensor
            Ground truth high resolution spatiotemporal data.
        hi_res_gen : tf.Tensor
            Superresolved high resolution spatiotemporal data generated by the
            generative model.
        weight_gen_advers : float
            Weight factor for the adversarial loss component of the generator
            vs. the discriminator.
        train_gen : bool
            True if generator is being trained, then loss=loss_gen
        train_disc : bool
            True if disc is being trained, then loss=loss_disc

        Returns
        -------
        loss : tf.Tensor
            0D tensor representing the loss value for the network being trained
            (either generator or one of the discriminators)
        loss_details : dict
            Namespace of the breakdown of loss components
        """

        if hi_res_gen.shape != hi_res_true.shape:
            msg = ('The tensor shapes of the synthetic output {} and '
                   'true high res {} did not have matching shape! '
                   'Check the spatiotemporal enhancement multipliers in your '
                   'your model config and data handlers.'
                   .format(hi_res_gen.shape, hi_res_true.shape))
            logger.error(msg)
            raise RuntimeError(msg)

        msg = ('Special SolarCC model can only accept multi-day hourly '
               '(multiple of 24) high res data in the axis=3 position but '
               'received shape {}'.format(hi_res_true.shape))
        assert hi_res_true.shape[3] % 24 == 0

        msg = ('Special SolarCC model can only accept multi-day hourly '
               '(multiple of 24) high res synthetic data in the axis=3 '
               'position but received shape {}'.format(hi_res_gen.shape))
        assert hi_res_gen.shape[3] % 24 == 0

        t_len = hi_res_true.shape[3]
        n_days = int(t_len // 24)
        day_slices = [slice(self.STARTING_HOUR + x,
                            self.STARTING_HOUR + x + self.DAYLIGHT_HOURS)
                      for x in range(0, 24 * n_days, 24)]

        # sample only daylight hours for disc training and gen content loss
        disc_out_true = []
        disc_out_gen = []
        loss_gen_content = 0.0
        for tslice in day_slices:
            disc_t = self._tf_discriminate(hi_res_true[:, :, :, tslice, :])
            gen_c = self.calc_loss_gen_content(hi_res_true[:, :, :, tslice, :],
                                               hi_res_gen[:, :, :, tslice, :])
            disc_out_true.append(disc_t)
            loss_gen_content += gen_c

        # strided sampling of all hours for gen adverserial loss
        time_samples = range(0, t_len - self.STRIDE_LEN, self.STRIDE_LEN)
        time_samples = [slice(i0, i0 + self.DAYLIGHT_HOURS)
                        for i0 in time_samples]
        for t_slice in time_samples:
            disc_g = self._tf_discriminate(hi_res_gen[:, :, :, t_slice, :])
            disc_out_gen.append(disc_g)

        disc_out_true = tf.concat([disc_out_true], axis=0)
        disc_out_gen = tf.concat([disc_out_gen], axis=0)
        loss_disc = self.calc_loss_disc(disc_out_true, disc_out_gen)

        loss_gen_content /= len(day_slices)
        loss_gen_advers = self.calc_loss_gen_advers(disc_out_gen)
        loss_gen = (loss_gen_content + weight_gen_advers * loss_gen_advers)

        loss = None
        if train_gen:
            loss = loss_gen
        elif train_disc:
            loss = loss_disc

        loss_details = {'loss_gen': loss_gen,
                        'loss_gen_content': loss_gen_content,
                        'loss_gen_advers': loss_gen_advers,
                        'loss_disc': loss_disc,
                        }

        return loss, loss_details



# -*- coding: utf-8 -*-
"""
Sup3r feature handling module.

@author: bbenton
"""

from abc import ABC, abstractmethod
from collections import defaultdict
from concurrent.futures import as_completed
import logging
import numpy as np
import re
import xarray as xr

from rex import Resource
from rex.utilities.execution import SpawnProcessPool
from sup3r.utilities.utilities import (invert_pot_temp, invert_uv,
                                       rotor_equiv_ws,
                                       transform_rotate_wind,
                                       bvf_squared,
                                       get_raster_shape,
                                       inverse_mo_length
                                       )


np.random.seed(42)

logger = logging.getLogger(__name__)


class DerivedFeature(ABC):
    """Abstract class for special features which need to be derived from raw
    features"""

    @classmethod
    @abstractmethod
    def inputs(cls, feature):
        """Required inputs for derived feature"""

    @classmethod
    @abstractmethod
    def compute(cls, data, height):
        """Compute method for derived feature"""


class ClearSkyRatioH5(DerivedFeature):
    """Clear Sky Ratio feature class for computing from H5 data"""

    @classmethod
    def inputs(cls, feature):
        """Get list of raw features used in calculation of the clearsky ratio

        Parameters
        ----------
        feature : str
            Clearsky ratio feature name, needs to be "clearsky_ratio"

        Returns
        -------
        list
            List of required features for clearsky_ratio: clearsky_ghi, ghi
        """
        assert feature == 'clearsky_ratio'
        return ['clearsky_ghi', 'ghi']

    @classmethod
    def compute(cls, data, height=None):
        """Compute the clearsky ratio

        Parameters
        ----------
        data : dict
            dictionary of feature arrays used for this compuation, must include
            clearsky_ghi and ghi
        height : str | int
            Placeholder to match interface with other compute methods

        Returns
        -------
        cs_ratio : ndarray
            Clearsky ratio, e.g. the all-sky ghi / the clearsky ghi. NaN where
            nighttime.
        """

        # need to use a nightime threshold of 1 W/m2 because cs_ghi is stored
        # in integer format and weird binning patterns happen in the clearsky
        # ratio and cloud mask between 0 and 1 W/m2 and sunrise/sunset
        night_mask = data['clearsky_ghi'] <= 1

        # set any timestep with any nighttime equal to NaN to avoid weird
        # sunrise/sunset artifacts.
        night_mask = night_mask.any(axis=(0, 1))
        data['clearsky_ghi'][..., night_mask] = np.nan

        cs_ratio = data['ghi'] / data['clearsky_ghi']
        cs_ratio = cs_ratio.astype(np.float32)
        return cs_ratio


class ClearSkyRatioCC(DerivedFeature):
    """Clear Sky Ratio feature class for computing from climate change netcdf
    data"""

    @classmethod
    def inputs(cls, feature):
        """Get list of raw features used in calculation of the clearsky ratio

        Parameters
        ----------
        feature : str
            Clearsky ratio feature name, needs to be "clearsky_ratio"

        Returns
        -------
        list
            List of required features for clearsky_ratio: clearsky_ghi, rsds
            (rsds==ghi for cc datasets)
        """
        assert feature == 'clearsky_ratio'
        return ['clearsky_ghi', 'rsds']

    @classmethod
    def compute(cls, data, height=None):
        """Compute the daily average climate change clearsky ratio

        Parameters
        ----------
        data : dict
            dictionary of feature arrays used for this compuation, must include
            clearsky_ghi and rsds (rsds==ghi for cc datasets)
        height : str | int
            Placeholder to match interface with other compute methods

        Returns
        -------
        cs_ratio : ndarray
            Clearsky ratio, e.g. the all-sky ghi / the clearsky ghi. This is
            assumed to be daily average data for climate change source data.
        """

        cs_ratio = data['rsds'] / data['clearsky_ghi']
        cs_ratio = np.minimum(cs_ratio, 1)
        cs_ratio = np.maximum(cs_ratio, 0)

        return cs_ratio


class CloudMaskH5(DerivedFeature):
    """Cloud Mask feature class for computing from H5 data"""

    @classmethod
    def inputs(cls, feature):
        """Get list of raw features used in calculation of the cloud mask
        Parameters
        ----------
        feature : str
            Cloud mask feature name, needs to be "cloud_mask"

        Returns
        -------
        list
            List of required features for cloud_mask: clearsky_ghi, ghi
        """
        assert feature == 'cloud_mask'
        return ['clearsky_ghi', 'ghi']

    @classmethod
    def compute(cls, data, height=None):
        """Compute the cloud mask

        Parameters
        ----------
        data : dict
            dictionary of feature arrays used for this compuation, must include
            clearsky_ghi and ghi
        height : str | int
            Placeholder to match interface with other compute methods

        Returns
        -------
        cloud_mask : ndarray
            Cloud mask, e.g. 1 where cloudy, 0 where clear. NaN where
            nighttime. Data is float32 so it can be normalized without any
            integer weirdness.
        """

        # need to use a nightime threshold of 1 W/m2 because cs_ghi is stored
        # in integer format and weird binning patterns happen in the clearsky
        # ratio and cloud mask between 0 and 1 W/m2 and sunrise/sunset
        night_mask = data['clearsky_ghi'] <= 1

        # set any timestep with any nighttime equal to NaN to avoid weird
        # sunrise/sunset artifacts.
        night_mask = night_mask.any(axis=(0, 1))

        cloud_mask = data['ghi'] < data['clearsky_ghi']
        cloud_mask = cloud_mask.astype(np.float32)
        cloud_mask[night_mask] = np.nan
        cloud_mask = cloud_mask.astype(np.float32)
        return cloud_mask


class PotentialTempNC(DerivedFeature):
    """Potential Temperature feature class for NETCDF data. Needed since T is
    perturbation potential temperature."""

    @classmethod
    def inputs(cls, feature):
        height = Feature.get_height(feature)
        features = [f'T_{height}m']
        return features

    @classmethod
    def compute(cls, data, height):
        """Method to compute Potential Temperature from NETCDF data

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Height at which to compute the derived feature

        Returns
        -------
        ndarray
            Derived feature array

        """
        return data[f'T_{height}m'] + 300


class TempNC(DerivedFeature):
    """Temperature feature class for NETCDF data. Needed since T is potential
    temperature not standard temp."""

    @classmethod
    def inputs(cls, feature):
        height = Feature.get_height(feature)
        features = [f'PotentialTemp_{height}m',
                    f'Pressure_{height}m']
        return features

    @classmethod
    def compute(cls, data, height):
        """Method to compute T from NETCDF data

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Height at which to compute the derived feature

        Returns
        -------
        ndarray
            Derived feature array

        """
        return invert_pot_temp(data[f'PotentialTemp_{height}m'],
                               data[f'Pressure_{height}m'])


class PressureNC(DerivedFeature):
    """Pressure feature class for NETCDF data. Needed since P is perturbation
    pressure."""

    @classmethod
    def inputs(cls, feature):
        height = Feature.get_height(feature)
        features = [f'P_{height}m',
                    f'PB_{height}m']
        return features

    @classmethod
    def compute(cls, data, height):
        """Method to compute pressure from NETCDF data

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Height at which to compute the derived feature

        Returns
        -------
        ndarray
            Derived feature array

        """
        return data[f'P_{height}m'] + data[f'PB_{height}m']


class BVFreqSquaredNC(DerivedFeature):
    """BVF Squared feature class with needed inputs method and compute
    method"""

    @classmethod
    def inputs(cls, feature):
        height = Feature.get_height(feature)
        features = [f'PT_{height}m',
                    f'PT_{int(height) - 100}m']

        return features

    @classmethod
    def compute(cls, data, height):
        """Method to compute BVF squared from NETCDF data

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Height at which to compute the derived feature

        Returns
        -------
        ndarray
            Derived feature array

        """
        # T is perturbation potential temperature for wrf and the
        # base potential temperature is 300K
        bvf2 = np.float32(9.81 / 100)
        bvf2 *= (data[f'PT_{height}m'] - data[f'PT_{int(height) - 100}m'])
        bvf2 /= (data[f'PT_{height}m'] + data[f'PT_{int(height) - 100}m'])
        bvf2 /= np.float32(2)
        return bvf2


class InverseMonNC(DerivedFeature):
    """Inverse MO feature class with needed inputs method and compute method"""

    @classmethod
    def inputs(cls, feature):
        """Required inputs for inverse MO from NETCDF data

        Parameters
        ----------
        feature : str
            raw feature name. e.g. RMOL

        Returns
        -------
        list
            List of required features for computing RMOL
        """

        assert feature == 'RMOL'
        features = ['UST', 'HFX']
        return features

    @classmethod
    def compute(cls, data, height=None):
        """Method to compute Inverse MO from NC data

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Placeholder to match interface with other compute methods

        Returns
        -------
        ndarray
            Derived feature array

        """
        return inverse_mo_length(data['UST'], data['HFX'])


class BVFreqMon(DerivedFeature):
    """BVF MO feature class with needed inputs method and compute method"""

    @classmethod
    def inputs(cls, feature):
        """Required inputs for computing BVF times inverse MO from data

        Parameters
        ----------
        feature : str
            raw feature name. e.g. BVF_MO_100m

        Returns
        -------
        list
            List of required features for computing BVF_MO
        """
        height = Feature.get_height(feature)
        features = [f'BVF2_{height}m', 'RMOL']
        return features

    @classmethod
    def compute(cls, data, height):
        """Method to compute BVF MO from data

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Height at which to compute the derived feature

        Returns
        -------
        ndarray
            Derived feature array

        """
        bvf_mo = data[f'BVF2_{height}m']
        mask = data['RMOL'] != 0
        bvf_mo[mask] = bvf_mo[mask] / data['RMOL'][mask]

        # making this zero when not both bvf and mo are negative
        bvf_mo[data['RMOL'] >= 0] = 0
        bvf_mo[bvf_mo < 0] = 0

        return bvf_mo


class BVFreqSquaredH5(DerivedFeature):
    """BVF Squared feature class with needed inputs method and compute
    method"""

    @classmethod
    def inputs(cls, feature):
        """Required inputs for computing BVF squared

        Parameters
        ----------
        feature : str
            raw feature name. e.g. BVF2_100m

        Returns
        -------
        list
            List of required features for computing BVF2
        """
        height = Feature.get_height(feature)
        features = [f'temperature_{height}m',
                    f'temperature_{int(height) - 100}m',
                    f'pressure_{height}m',
                    f'pressure_{int(height) - 100}m']

        return features

    @classmethod
    def compute(cls, data, height):
        """Method to compute BVF squared from H5 data

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Height at which to compute the derived feature

        Returns
        -------
        ndarray
            Derived feature array

        """
        return bvf_squared(
            data[f'temperature_{height}m'],
            data[f'temperature_{int(height) - 100}m'],
            data[f'pressure_{height}m'],
            data[f'pressure_{int(height) - 100}m'],
            100)


class WindspeedNC(DerivedFeature):
    """Windspeed feature from netcdf data"""

    @classmethod
    def inputs(cls, feature):
        """Required inputs for computing windspeed from netcdf data

        Parameters
        ----------
        feature : str
            raw feature name. e.g. BVF_MO_100m

        Returns
        -------
        list
            List of required features for computing windspeed
        """

        height = Feature.get_height(feature)
        features = [f'U_{height}m', f'V_{height}m', 'lat_lon']
        return features

    @classmethod
    def compute(cls, data, height):
        """Compute windspeed

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Height at which to compute the derived feature

        Returns
        -------
        ndarray
            Derived feature array
        """

        ws, _ = invert_uv(data[f'U_{height}m'], data[f'V_{height}m'],
                          data['lat_lon'])
        return ws


class WinddirectionNC(DerivedFeature):
    """Winddirection feature from netcdf data"""

    @classmethod
    def inputs(cls, feature):
        """Required inputs for computing windspeed from netcdf data

        Parameters
        ----------
        feature : str
            raw feature name. e.g. BVF_MO_100m

        Returns
        -------
        list
            List of required features for computing windspeed
        """

        height = Feature.get_height(feature)
        features = [f'U_{height}m', f'V_{height}m', 'lat_lon']
        return features

    @classmethod
    def compute(cls, data, height):
        """Compute winddirection

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Height at which to compute the derived feature

        Returns
        -------
        ndarray
            Derived feature array
        """

        _, wd = invert_uv(data[f'U_{height}m'], data[f'V_{height}m'],
                          data['lat_lon'])
        return wd


class Veer(DerivedFeature):
    """Veer at a given height"""

    HEIGHTS = [40, 60, 80, 100, 120]

    @classmethod
    def inputs(cls, feature):
        """Required inputs for computing Veer

        Parameters
        ----------
        feature : str
            raw feature name. e.g. BVF_MO_100m

        Returns
        -------
        list
            List of required features for computing REWS
        """

        rotor_center = Feature.get_height(feature)
        if rotor_center is None:
            heights = cls.HEIGHTS
        else:
            heights = [int(rotor_center) - i * 20 for i in [-2, -1, 0, 1, 2]]
        features = []
        for height in heights:
            features.append(f'winddirection_{height}m')
        return features

    @classmethod
    def compute(cls, data, height):
        """Compute Veer

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Height at which to compute the derived feature

        Returns
        -------
        ndarray
            Derived feature array
        """

        if height is None:
            heights = cls.HEIGHTS
        else:
            heights = [int(height) - i * 20 for i in [-2, -1, 0, 1, 2]]
        veer = 0
        for i in range(0, len(heights), 2):
            tmp = np.radians(data[f'winddirection_{height[i + 1]}'])
            tmp -= np.radians(data[f'winddirection_{height[i]}'])
            veer += np.abs(tmp)
        veer /= (heights[-1] - heights[0])
        return veer


class Shear(DerivedFeature):
    """Wind shear at a given height"""

    @classmethod
    def inputs(cls, feature):
        """Required inputs for computing Veer

        Parameters
        ----------
        feature : str
            raw feature name. e.g. BVF_MO_100m

        Returns
        -------
        list
            List of required features for computing Veer
        """

        height = Feature.get_height(feature)
        heights = [int(height), int(height) + 20]
        features = []
        for height in heights:
            features.append(f'winddirection_{height}m')
        return features

    @classmethod
    def compute(cls, data, height):
        """Compute REWS

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Height at which to compute the derived feature

        Returns
        -------
        ndarray
            Derived feature array
        """

        heights = [int(height), int(height) + 20]
        shear = np.cos(np.radians(data[f'winddirection_{int(height) + 20}m']))
        shear -= np.cos(np.radians(data[f'winddirection_{int(height)}m']))
        shear /= (heights[-1] - heights[0])
        return shear


class Rews(DerivedFeature):
    """Rotor equivalent wind speed"""

    HEIGHTS = [40, 60, 80, 100, 120]

    @classmethod
    def inputs(cls, feature):
        """Required inputs for computing REWS

        Parameters
        ----------
        feature : str
            raw feature name. e.g. BVF_MO_100m

        Returns
        -------
        list
            List of required features for computing REWS
        """

        rotor_center = Feature.get_height(feature)
        if rotor_center is None:
            heights = cls.HEIGHTS
        else:
            heights = [int(rotor_center) - i * 20 for i in [-2, -1, 0, 1, 2]]
        features = []
        for height in heights:
            features.append(f'windspeed_{height}m')
            features.append(f'winddirection_{height}m')
        return features

    @classmethod
    def compute(cls, data, height):
        """Compute REWS

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Height at which to compute the derived feature

        Returns
        -------
        ndarray
            Derived feature array
        """

        if height is None:
            heights = cls.HEIGHTS
        else:
            heights = [int(height) - i * 20 for i in [-2, -1, 0, 1, 2]]
        rews = rotor_equiv_ws(data, heights)
        return rews


class UWind(DerivedFeature):
    """U wind component feature class with needed inputs method and compute
    method"""

    @classmethod
    def inputs(cls, feature):
        """Required inputs for computing U wind component

        Parameters
        ----------
        feature : str
            raw feature name. e.g. U_100m

        Returns
        -------
        list
            List of required features for computing U
        """
        height = Feature.get_height(feature)
        features = [f'windspeed_{height}m', f'winddirection_{height}m',
                    'lat_lon']
        return features

    @classmethod
    def compute(cls, data, height):
        """Method to compute U wind component from H5 data

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Height at which to compute the derived feature

        Returns
        -------
        ndarray
            Derived feature array

        """
        u, _ = transform_rotate_wind(data[f'windspeed_{height}m'],
                                     data[f'winddirection_{height}m'],
                                     data['lat_lon'])
        return u


class VWind(DerivedFeature):
    """V wind component feature class with needed inputs method and compute
    method"""

    @classmethod
    def inputs(cls, feature):
        """Required inputs for computing V wind component

        Parameters
        ----------
        feature : str
            raw feature name. e.g. V_100m

        Returns
        -------
        list
            List of required features for computing V
        """
        height = Feature.get_height(feature)
        features = [f'windspeed_{height}m', f'winddirection_{height}m',
                    'lat_lon']
        return features

    @classmethod
    def compute(cls, data, height):
        """Method to compute V wind component from H5 data

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Height at which to compute the derived feature

        Returns
        -------
        ndarray
            Derived feature array

        """
        _, v = transform_rotate_wind(data[f'windspeed_{height}m'],
                                     data[f'winddirection_{height}m'],
                                     data['lat_lon'])
        return v


class Tas(DerivedFeature):
    """Air temperature near surface variable from climate change nc files"""

    @classmethod
    def inputs(cls, feature):
        """Required inputs for computing tas

        Parameters
        ----------
        feature : str
            raw feature name. e.g. tas

        Returns
        -------
        list
            List of required features for computing tas
        """
        return ['tas']

    @classmethod
    def compute(cls, data, height):
        """Method to compute tas in Celsius from tas source in Kelvin

        Parameters
        ----------
        data : dict
            Dictionary of raw feature arrays to use for derivation
        height : str | int
            Height at which to compute the derived feature

        Returns
        -------
        ndarray
            Derived feature array
        """
        return data['tas'] - 273.15


class LatLonNC:
    """Lat Lon feature class with compute method"""

    @classmethod
    def compute(cls, file_paths, raster_index):
        """Get lats and lons

        Parameters
        ----------
        file_paths : list
            path to data file
        raster_index : list
            List of slices for raster

        Returns
        -------
        ndarray
            lat lon array
            (spatial_1, spatial_2, 2)
        """
        with xr.open_dataset(file_paths[0]) as handle:
            lat_key = 'XLAT'
            lon_key = 'XLONG'
            if lat_key not in handle.variables:
                lat_key = 'latitude'
            if lon_key not in handle.variables:
                lon_key = 'longitude'
            if len(handle.variables[lat_key].dims) == 3:
                idx = (0, raster_index[0], raster_index[1])
            elif len(handle.variables[lat_key].dims) == 4:
                idx = (0, raster_index[0], raster_index[1], 0)
            else:
                idx = (raster_index[0], raster_index[1])
            lats = handle.variables[lat_key].values[idx]
            lons = handle.variables[lon_key].values[idx]
            lat_lon = np.dstack((lats, lons))
        return lat_lon


class LatLonNCforCC:
    """Lat Lon feature class with compute method"""

    @classmethod
    def compute(cls, file_paths, raster_index):
        """Get lats and lons

        Parameters
        ----------
        file_paths : list
            path to data file
        raster_index : list
            List of slices for raster

        Returns
        -------
        ndarray
            lat lon array
            (spatial_1, spatial_2, 2)
        """
        with xr.open_dataset(file_paths[0]) as handle:
            lats = (handle.lat.values if 'time' not
                    in handle.lat.dims else handle.lat.values[0])
            lons = (handle.lon.values if 'time' not
                    in handle.lon.dims else handle.lon.values[0])
            lons, lats = np.meshgrid(lons, lats)
            lat_lon = np.dstack((lats[tuple(raster_index)],
                                 lons[tuple(raster_index)]))
        return lat_lon


class TopoH5:
    """Topography feature class with compute method"""

    @classmethod
    def compute(cls, file_paths, raster_index):
        """Get topography corresponding to raster

        Parameters
        ----------
        file_paths : list
            path to data file
        raster_index : ndarray
            Raster index array

        Returns
        -------
        ndarray
            topo array
            (spatial_1, spatial_2)
        """
        with Resource(file_paths[0], hsds=False) as handle:
            idx = tuple([raster_index.flatten()])
            topo = handle.get_meta_arr('elevation')[idx]
            topo = topo.reshape((raster_index.shape[0], raster_index.shape[1]))
        return topo


class LatLonH5:
    """Lat Lon feature class with compute method"""

    @classmethod
    def compute(cls, file_paths, raster_index):
        """Get lats and lons corresponding to raster for use in
        windspeed/direction -> u/v mapping

        Parameters
        ----------
        file_paths : list
            path to data file
        raster_index : ndarray
            Raster index array

        Returns
        -------
        ndarray
            lat lon array
            (spatial_1, spatial_2, 2)
        """
        with Resource(file_paths[0], hsds=False) as handle:
            lat_lon = handle.lat_lon[tuple([raster_index.flatten()])]
            lat_lon = lat_lon.reshape((raster_index.shape[0],
                                       raster_index.shape[1], 2))
        return lat_lon


class Feature:
    """Class to simplify feature computations. Stores feature height, feature
    basename, name of feature in handle"""

    def __init__(self, feature, handle):
        """Takes a feature (e.g. U_100m) and gets the height (100), basename
        (U) and determines whether the feature is found in the data handle

        Parameters
        ----------
        feature : str
            Raw feature name e.g. U_100m
        handle : WindX | NSRDBX | xarray
            handle for data file
        """
        self.raw_name = feature
        self.height = self.get_height(feature)
        self.pressure = self.get_pressure(feature)
        self.basename = self.get_basename(feature)
        if self.raw_name in handle:
            self.handle_input = self.raw_name
        elif self.basename in handle:
            self.handle_input = self.basename
        else:
            self.handle_input = None

    @staticmethod
    def get_basename(feature):
        """Get basename of feature. e.g. temperature from temperature_100m

        Parameters
        ----------
        feature : str
            Name of feature. e.g. U_100m

        Returns
        -------
        str
            feature basename
        """

        height = Feature.get_height(feature)
        pressure = Feature.get_pressure(feature)
        if height is not None or pressure is not None:
            suffix = feature.split('_')[-1]
            basename = feature.replace(f'_{suffix}', '')
        else:
            basename = feature
        return basename

    @staticmethod
    def get_height(feature):
        """Get height from feature name to use in height interpolation

        Parameters
        ----------
        feature : str
            Name of feature. e.g. U_100m

        Returns
        -------
        float | None
            height to use for interpolation
            in meters
        """
        height = re.search(r'\d+m', feature)
        if height:
            height = height.group(0).strip('m')
            if not height.isdigit():
                height = None
        return height

    @staticmethod
    def get_pressure(feature):
        """Get pressure from feature name to use in pressure interpolation

        Parameters
        ----------
        feature : str
            Name of feature. e.g. U_100pa

        Returns
        -------
        float | None
            pressure to use for interpolation in pascals
        """
        pressure = re.search(r'\d+pa', feature)
        if pressure:
            pressure = pressure.group(0).strip('pa')
            if not pressure.isdigit():
                pressure = None
        return pressure


class FeatureHandler:
    """Feature Handler with cache for previously loaded features used in other
    calculations """

    @classmethod
    def valid_handle_features(cls, features, handle_features):
        """Check if features are in handle

        Parameters
        ----------
        features : str | list
            Raw feature names e.g. U_100m
        handle_features : list
            Features available in raw data

        Returns
        -------
        bool
            Whether feature basename is in handle
        """

        if features is None:
            return False

        return all(Feature.get_basename(f) in handle_features
                   or f in handle_features for f in features)

    @classmethod
    def valid_input_features(cls, features, handle_features):
        """Check if features are in handle or have compute methods

        Parameters
        ----------
        features : str | list
            Raw feature names e.g. U_100m
        handle_features : list
            Features available in raw data

        Returns
        -------
        bool
            Whether feature basename is in handle
        """

        if features is None:
            return False

        if all(Feature.get_basename(f) in handle_features
               or f in handle_features
               or cls.lookup(f, 'compute') is not None for f in features):
            return True
        return False

    @classmethod
    def pop_old_data(cls, data, chunk_number, all_features):
        """Remove input feature data if no longer needed for requested features

        Parameters
        ----------
        data : dict
            dictionary of feature arrays with integer keys for chunks and str
            keys for features.  e.g. data[chunk_number][feature] = array.
            (spatial_1, spatial_2, temporal)
        chunk_number : int
            time chunk index to check
        all_features : list
            list of all requested features including those requiring derivation
            from input features

        """
        if data:
            old_keys = [f for f in data[chunk_number]
                        if f not in all_features]
            for k in old_keys:
                data[chunk_number].pop(k)

    @classmethod
    def serial_extract(cls, file_paths, raster_index, time_chunks,
                       input_features):
        """Extract features in series

        Parameters
        ----------
        file_paths : list
            list of file paths
        raster_index : ndarray
            raster index for spatial domain
        time_chunks : list
            List of slices to chunk data feature extraction along time
            dimension
        input_features : list
            list of input feature strings

        Returns
        -------
        dict
            dictionary of feature arrays with integer keys for chunks and str
            keys for features.  e.g. data[chunk_number][feature] = array.
            (spatial_1, spatial_2, temporal)
        """

        data = defaultdict(dict)
        for t, t_slice in enumerate(time_chunks):
            for f in input_features:
                data[t][f] = cls.extract_feature(file_paths, raster_index, f,
                                                 t_slice)
            interval = np.int(np.ceil(len(time_chunks) / 10))
            if interval > 0 and t % interval == 0:
                logger.debug(f'{t+1} out of {len(time_chunks)} feature '
                             'chunks extracted.')
        return data

    @classmethod
    def parallel_extract(cls, file_paths, raster_index, time_chunks,
                         input_features, max_workers=None):
        """Extract features using parallel subprocesses

        Parameters
        ----------
        file_paths : list
            list of file paths
        raster_index : ndarray | list
            raster index for spatial domain
        time_chunks : list
            List of slices to chunk data feature extraction along time
            dimension
        input_features : list
            list of input feature strings
        max_workers : int | None
            Number of max workers to use for extraction.  If equal to 1 then
            method is run in serial

        Returns
        -------
        dict
            dictionary of feature arrays with integer keys for chunks and str
            keys for features.  e.g. data[chunk_number][feature] = array.
            (spatial_1, spatial_2, temporal)
        """
        futures = {}
        data = defaultdict(dict)

        with SpawnProcessPool(max_workers=max_workers) as exe:
            for t, t_slice in enumerate(time_chunks):
                for f in input_features:
                    future = exe.submit(cls.extract_feature,
                                        file_paths=file_paths,
                                        raster_index=raster_index,
                                        feature=f,
                                        time_slice=t_slice)
                    meta = {'feature': f, 'chunk': t}
                    futures[future] = meta

            shape = get_raster_shape(raster_index)
            time_shape = time_chunks[0].stop - time_chunks[0].start
            time_shape //= time_chunks[0].step
            logger.info(f'Started extracting {input_features}'
                        f' using {len(time_chunks)}'
                        f' time chunks of shape ({shape[0]}, {shape[1]}, '
                        f'{time_shape}) for {len(input_features)} features')

            interval = np.int(np.ceil(len(futures) / 10))
            for i, future in enumerate(as_completed(futures)):
                v = futures[future]
                try:
                    data[v['chunk']][v['feature']] = future.result()
                except Exception as e:
                    msg = (f'Error extracting chunk {v["chunk"]} for'
                           f' {v["feature"]}')
                    logger.error(msg)
                    raise RuntimeError(msg) from e
                if interval > 0 and i % interval == 0:
                    logger.debug(f'{i+1} out of {len(futures)} feature '
                                 'chunks extracted.')

        return data

    @classmethod
    def recursive_compute(cls, data, feature, handle_features, file_paths,
                          raster_index):
        """Compute intermediate features recursively

        Parameters
        ----------
        data : dict
            dictionary of feature arrays. e.g. data[feature] = array.
            (spatial_1, spatial_2, temporal)
        feature : str
            Name of feature to compute
        handle_features : list
            Features available in raw data
        file_paths : list
            Paths to data files. Used if compute method operates directly on
            source handler instead of input arrays. This is done with features
            without inputs methods like lat_lon and topography.
        raster_index : ndarray
            raster index for spatial domain

        Returns
        -------
        ndarray
            Array of computed feature data
        """
        if feature not in data:
            inputs = cls.lookup(feature, 'inputs',
                                handle_features=handle_features)
            method = cls.lookup(feature, 'compute')
            height = Feature.get_height(feature)
            if inputs is not None:
                if method is None:
                    return data[inputs(feature)[0]]
                elif all(r in data for r in inputs(feature)):
                    data[feature] = method(data, height)
                else:
                    for r in inputs(feature):
                        data[r] = cls.recursive_compute(data, r,
                                                        handle_features,
                                                        file_paths,
                                                        raster_index)
                    data[feature] = method(data, height)
            elif method is not None:
                data[feature] = method(file_paths, raster_index)

        return data[feature]

    @classmethod
    def serial_compute(cls, data, file_paths, raster_index, time_chunks,
                       derived_features, all_features, handle_features):
        """Compute features in series

        Parameters
        ----------
        data : dict
            dictionary of feature arrays with integer keys for chunks and str
            keys for features. e.g. data[chunk_number][feature] = array.
            (spatial_1, spatial_2, temporal)
        file_paths : list
            Paths to data files. Used if compute method operates directly on
            source handler instead of input arrays. This is done with features
            without inputs methods like lat_lon and topography.
        raster_index : ndarray
            raster index for spatial domain
        time_chunks : list
            List of slices to chunk data feature extraction along time
            dimension
        derived_features : list
            list of feature strings which need to be derived
        all_features : list
            list of all features including those requiring derivation from
            input features
        handle_features : list
            Features available in raw data

        Returns
        -------
        data : dict
            dictionary of feature arrays, including computed features, with
            integer keys for chunks and str keys for features.
            e.g. data[chunk_number][feature] = array.
            (spatial_1, spatial_2, temporal)
        """

        if len(derived_features) == 0:
            return data

        for t, _ in enumerate(time_chunks):
            for _, f in enumerate(derived_features):
                tmp = cls.get_input_arrays(data, t, f, handle_features)
                data[t][f] = cls.recursive_compute(
                    data=tmp, feature=f, handle_features=handle_features,
                    file_paths=file_paths, raster_index=raster_index)
            cls.pop_old_data(data, t, all_features)
            interval = np.int(np.ceil(len(time_chunks) / 10))
            if interval > 0 and t % interval == 0:
                logger.debug(f'{t+1} out of {len(time_chunks)} feature '
                             'chunks computed.')

        return data

    @classmethod
    def parallel_compute(cls, data, file_paths, raster_index, time_chunks,
                         derived_features, all_features, handle_features,
                         max_workers=None):
        """Compute features using parallel subprocesses

        Parameters
        ----------
        data : dict
            dictionary of feature arrays with integer keys for chunks and str
            keys for features.
            e.g. data[chunk_number][feature] = array.
            (spatial_1, spatial_2, temporal)
        file_paths : list
            Paths to data files. Used if compute method operates directly on
            source handler instead of input arrays. This is done with features
            without inputs methods like lat_lon and topography.
        raster_index : ndarray
            raster index for spatial domain
        time_chunks : list
            List of slices to chunk data feature extraction along time
            dimension
        derived_features : list
            list of feature strings which need to be derived
        all_features : list
            list of all features including those requiring derivation from
            input features
        handle_features : list
            Features available in raw data
        max_workers : int | None
            Number of max workers to use for computation. If equal to 1 then
            method is run in serial

        Returns
        -------
        data : dict
            dictionary of feature arrays, including computed features, with
            integer keys for chunks and str keys for features. Includes e.g.
            data[chunk_number][feature] = array.
            (spatial_1, spatial_2, temporal)
        """
        if len(derived_features) == 0:
            return data

        futures = {}
        with SpawnProcessPool(max_workers=max_workers) as exe:
            for t, _ in enumerate(time_chunks):
                for f in derived_features:
                    tmp = cls.get_input_arrays(data, t, f, handle_features)
                    future = exe.submit(cls.recursive_compute, data=tmp,
                                        feature=f,
                                        handle_features=handle_features,
                                        file_paths=file_paths,
                                        raster_index=raster_index)
                    meta = {'feature': f, 'chunk': t}
                    futures[future] = meta

                cls.pop_old_data(data, t, all_features)

            shape = get_raster_shape(raster_index)
            time_shape = time_chunks[0].stop - time_chunks[0].start
            time_shape //= time_chunks[0].step
            logger.info(f'Started computing {derived_features}'
                        f' using {len(time_chunks)}'
                        f' time chunks of shape ({shape[0]}, {shape[1]}, '
                        f'{time_shape}) for {len(derived_features)} features')

            interval = np.int(np.ceil(len(futures) / 10))
            for i, future in enumerate(as_completed(futures)):
                v = futures[future]
                chunk_idx = v['chunk']
                data[chunk_idx] = data.get(chunk_idx, {})
                data[chunk_idx][v['feature']] = future.result()
                if interval > 0 and i % interval == 0:
                    logger.debug(f'{i+1} out of {len(futures)} feature '
                                 'chunks computed')

        return data

    @classmethod
    def get_input_arrays(cls, data, chunk_number, f, handle_features):
        """Get only arrays needed for computations

        Parameters
        ----------
        data : dict
            Dictionary of feature arrays
        chunk_number :
            time chunk for which to get input arrays
        f : str
            feature to compute using input arrays
        handle_features : list
            Features available in raw data

        Returns
        -------
        dict
            Dictionary of arrays with only needed features
        """
        tmp = {}
        if data:
            inputs = cls.get_inputs_recursive(f, handle_features)
            for r in inputs:
                if r in data[chunk_number]:
                    tmp[r] = data[chunk_number][r]
        return tmp

    @classmethod
    def lookup(cls, feature, attr_name, handle_features=None):
        """Lookup feature in feature registry

        Parameters
        ----------
        feature : str
            Feature to lookup in registry
        attr_name : str
            Type of method to lookup. e.g. inputs or compute
        handle_features : list
            List of feature names (datasets) available in the source file. If
            feature is found explicitly in this list, height/pressure suffixes
            will not be appended to the output.

        Returns
        -------
        method | None
            Feature registry method corresponding to feature
        """

        if handle_features is None:
            handle_features = []

        out = None
        for k, v in cls.feature_registry().items():
            if re.match(k.lower(), feature.lower()):
                out = v
                break

        if out is None:
            return None

        if not isinstance(out, str):
            return getattr(out, attr_name, None)

        elif attr_name == 'inputs':

            if out in handle_features:
                return lambda x: [out]

            height = Feature.get_height(feature)
            if height is not None:
                out = out.split('(.*)')[0] + f'{height}m'

            pressure = Feature.get_pressure(feature)
            if pressure is not None:
                out = out.split('(.*)')[0] + f'{pressure}pa'

            return lambda x: [out]

    @classmethod
    def get_inputs_recursive(cls, feature, handle_features):
        """Lookup inputs needed to compute feature. Walk through inputs methods
        for each required feature to get all raw features.

        Parameters
        ----------
        feature : str
            Feature for which to get needed inputs for derivation
        handle_features : list
            Features available in raw data

        Returns
        -------
        list
            List of input features
        """
        raw_features = []
        method = cls.lookup(feature, 'inputs', handle_features=handle_features)

        check1 = feature not in raw_features
        check2 = (cls.valid_handle_features([feature], handle_features)
                  or method is None)
        if check1 and check2:
            raw_features.append(feature)

        else:
            for f in method(feature):
                lkup = cls.lookup(f, 'inputs', handle_features=handle_features)
                valid = cls.valid_handle_features([f], handle_features)
                if (lkup is None or valid) and f not in raw_features:
                    raw_features.append(f)
                else:
                    for r in cls.get_inputs_recursive(f, handle_features):
                        if r not in raw_features:
                            raw_features.append(r)

        return raw_features

    @classmethod
    def get_raw_feature_list(cls, features, handle_features):
        """Lookup inputs needed to compute feature

        Parameters
        ----------
        features : list
            Features for which to get needed inputs for derivation
        handle_features : list
            Features available in raw data

        Returns
        -------
        list
            List of input features
        """
        raw_features = []
        for f in features:
            candidate_features = cls.get_inputs_recursive(f, handle_features)
            if candidate_features:
                for r in candidate_features:
                    if r not in raw_features:
                        raw_features.append(r)
            else:
                req = cls.lookup(f, "inputs", handle_features=handle_features)
                req = req(f)
                msg = (f'Cannot compute {f} from the provided data. '
                       f'Requested features: {req}')
                logger.error(msg)
                raise ValueError(msg)
        return raw_features

    @classmethod
    @abstractmethod
    def feature_registry(cls):
        """Registry of methods for computing features

        Returns
        -------
        dict
            Method registry
        """

    @classmethod
    @abstractmethod
    def extract_feature(cls, file_paths, raster_index, feature,
                        time_slice=slice(None)):
        """Extract single feature from data source

        Parameters
        ----------
        file_paths : list
            path to data file
        raster_index : ndarray
            Raster index array
        time_slice : slice
            slice of time to extract
        feature : str
            Feature to extract from data

        Returns
        -------
        ndarray
            Data array for extracted feature
            (spatial_1, spatial_2, temporal)
        """



# -*- coding: utf-8 -*-
"""Custom sup3r solar module. This primarily converts GAN output clearsky ratio
to GHI, DNI, and DHI using NSRDB data and utility modules like DISC

Note that clearsky_ratio is assumed to be clearsky ghi ratio and is calculated
as daily average GHI / daily average clearsky GHI.
"""
import glob
import json
import os
import numpy as np
import logging
from scipy.spatial import KDTree
from farms.disc import disc
from farms.utilities import calc_dhi, dark_night
from rex import Resource, MultiTimeResource
from rex.utilities.fun_utils import get_fun_call_str

from sup3r.utilities import ModuleName
from sup3r.postprocessing.file_handling import RexOutputs, H5_ATTRS


logger = logging.getLogger(__name__)


class Solar:
    """Custom sup3r solar module. This primarily converts GAN output clearsky
    ratio to GHI, DNI, and DHI using NSRDB data and utility modules like
    DISC"""

    def __init__(self, sup3r_fps, nsrdb_fp, t_slice=slice(None), tz=-6,
                 agg_factor=1, nn_threshold=0.5, cloud_threshold=0.99):
        """
        Parameters
        ----------
        sup3r_fps : str | list
            Full .h5 filepath(s) to one or more sup3r GAN output .h5 chunk
            files containing clearsky_ratio, time_index, and meta. These files
            must have the same meta data but can be sequential and ordered
            temporal chunks. The data in this file has been rolled by tz
            to a local time (assumed that the GAN was trained on local time
            solar data) and will be converted back into UTC, so it's wise to
            include some padding in the sup3r_fps file list.
        nsrdb_fp : str
            Filepath to NSRDB .h5 file containing clearsky_ghi, clearsky_dni,
            clearsky_dhi data.
        t_slice : slice
            Slicing argument to slice the temporal axis of the sup3r_fps source
            data after doing the tz roll to UTC but before returning the
            irradiance variables. This can be used to effectively pad the solar
            irradiance calculation in UTC time. For example, if sup3r_fps is 3
            files each with 24 hours of data, t_slice can be slice(24, 48) to
            only output the middle day of irradiance data, but padded by the
            other two days for the UTC output.
        tz : int
            The timezone offset for the data in sup3r_fps. It is assumed that
            the GAN is trained on data in local time and therefore the output
            in sup3r_fps should be treated as local time. For example, -6 is
            CST which is default for CONUS training data.
        agg_factor : int
            Spatial aggregation factor for nsrdb-to-GAN-meta e.g. the number of
            NSRDB spatial pixels to average for a single sup3r GAN output site.
        nn_threshold : float
            The KDTree nearest neighbor threshold that determines how far the
            sup3r GAN output data has to be from the NSRDB source data to get
            irradiance=0. Note that is value is in decimal degrees which is a
            very approximate way to determine real distance.
        cloud_threshold : float
            Clearsky ratio threshold below which the data is considered cloudy
            and DNI is calculated using DISC.
        """

        self.t_slice = t_slice
        self.agg_factor = agg_factor
        self.nn_threshold = nn_threshold
        self.cloud_threshold = cloud_threshold
        self.tz = tz
        self._nsrdb_fp = nsrdb_fp
        self._sup3r_fps = sup3r_fps
        if isinstance(self._sup3r_fps, str):
            self._sup3r_fps = [self._sup3r_fps]

        logger.debug('Initializing solar module with sup3r files: {}'
                     .format([os.path.basename(fp) for fp in self._sup3r_fps]))
        logger.debug('Initializing solar module with temporal slice: {}'
                     .format(self.t_slice))
        logger.debug('Initializing solar module with NSRDB source fp: {}'
                     .format(self._nsrdb_fp))

        self.gan_data = MultiTimeResource(self._sup3r_fps)
        self.nsrdb = Resource(self._nsrdb_fp)

        # cached variables
        self._nsrdb_tslice = None
        self._idnn = None
        self._dist = None
        self._sza = None
        self._cs_ratio = None
        self._ghi = None
        self._dni = None
        self._dhi = None

        self.preflight()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()
        if type is not None:
            raise

    def preflight(self):
        """Run preflight checks on source data to make sure everything will
        work together."""
        assert 'clearsky_ratio' in self.gan_data.dsets
        assert 'clearsky_ghi' in self.nsrdb.dsets
        assert 'clearsky_dni' in self.nsrdb.dsets
        assert 'solar_zenith_angle' in self.nsrdb.dsets
        assert 'surface_pressure' in self.nsrdb.dsets
        assert isinstance(self.nsrdb_tslice, slice)

        ti_gan = self.gan_data.time_index
        ti_gan_1 = np.roll(ti_gan, 1)
        delta = (ti_gan - ti_gan_1)[1:].mean().total_seconds()
        msg = ('Its assumed that the sup3r GAN output solar data will be '
               'hourly but received time index: {}'.format(ti_gan))
        assert delta == 3600, msg

    def close(self):
        """Close all internal file handlers"""
        self.gan_data.close()
        self.nsrdb.close()

    @property
    def idnn(self):
        """Get the nearest neighbor meta data indices from the NSRDB data that
        correspond to the sup3r GAN data

        Returns
        -------
        idnn : np.ndarray
            2D array of length (n_sup3r_sites, agg_factor) where the values are
            meta data indices from the NSRDB.
        """
        if self._idnn is None:
            tree = KDTree(self.nsrdb.meta[['latitude', 'longitude']])
            out = tree.query(self.gan_data.meta[['latitude', 'longitude']])
            self._dist, self._idnn = out
            if len(self._idnn.shape) == 1:
                self._dist = np.expand_dims(self._dist, axis=1)
                self._idnn = np.expand_dims(self._idnn, axis=1)
        return self._idnn

    @property
    def dist(self):
        """Get the nearest neighbor distances from the sup3r GAN data sites to
        the NSRDB nearest neighbors.

        Returns
        -------
        dist : np.ndarray
            2D array of length (n_sup3r_sites, agg_factor) where the values are
            decimal degree distances from the sup3r sites to the nsrdb nearest
            neighbors.
        """
        if self._dist is None:
            _ = self.idnn
        return self._dist

    @property
    def time_index(self):
        """Time index for the sup3r GAN output data but sliced by t_slice

        Returns
        -------
        pd.DatetimeIndex
        """
        return self.gan_data.time_index[self.t_slice]

    @property
    def out_of_bounds(self):
        """Get a boolean mask for the sup3r data that is out of bounds (too far
        from the NSRDB data).

        Returns
        -------
        out_of_bounds : np.ndarray
            1D boolean array with length == number of sup3r GAN sites. True if
            the site is too far from the NSRDB.
        """
        return (self.dist > self.nn_threshold).any(axis=1)

    @property
    def nsrdb_tslice(self):
        """Get the time slice of the NSRDB data corresponding to the sup3r GAN
        output."""

        if self._nsrdb_tslice is None:
            doy_nsrdb = self.nsrdb.time_index.day_of_year
            doy_gan = self.time_index.day_of_year
            mask = doy_nsrdb.isin(doy_gan)

            if mask.sum() == 0:
                msg = ('Time index intersection of the NSRDB time index and '
                       'sup3r GAN output has only {} common timesteps! '
                       'Something went wrong.\nNSRDB time index: \n{}\nSup3r '
                       'GAN output time index:\n{}'
                       .format(mask.sum(), self.nsrdb.time_index,
                               self.time_index))
                logger.error(msg)
                raise RuntimeError(msg)

            ilocs = np.where(mask)[0]
            t0, t1 = ilocs[0], ilocs[-1] + 1

            ti_nsrdb = self.nsrdb.time_index
            ti_nsrdb_1 = np.roll(ti_nsrdb, 1)
            delta = (ti_nsrdb - ti_nsrdb_1)[1:].mean().total_seconds()
            step = int(3600 // delta)
            self._nsrdb_tslice = slice(t0, t1, step)

            logger.debug('Found nsrdb_tslice {} with corresponding '
                         'time index:\n\t{}'
                         .format(self._nsrdb_tslice,
                                 self.nsrdb.time_index[self._nsrdb_tslice]))

        return self._nsrdb_tslice

    @property
    def clearsky_ratio(self):
        """Get the clearsky ghi ratio data from the GAN output, rolled from tz
        to UTC.

        Returns
        -------
        clearsky_ratio : np.ndarray
            2D array with shape (time, sites) in UTC.
        """
        if self._cs_ratio is None:
            self._cs_ratio = self.gan_data['clearsky_ratio']
            self._cs_ratio = np.roll(self._cs_ratio, -self.tz, axis=0)

            # if tz is negative, roll to utc is positive, and the beginning of
            # the dataset is rolled over from the end and must be backfilled,
            # otherwise you can get seams
            self._cs_ratio[:-self.tz, :] = self._cs_ratio[-self.tz, :]

            # apply temporal slicing of source data, see docstring on t_slice
            # for more info
            self._cs_ratio = self._cs_ratio[self.t_slice, :]

        return self._cs_ratio

    @property
    def solar_zenith_angle(self):
        """Get the solar zenith angle (degrees)

        Returns
        -------
        solar_zenith_angle : np.ndarray
            2D array with shape (time, sites) in UTC.
        """
        if self._sza is None:
            self._sza = self.get_nsrdb_data('solar_zenith_angle')
        return self._sza

    @property
    def ghi(self):
        """Get the ghi (W/m2) based on the GAN output clearsky ratio +
        clearsky_ghi in UTC.

        Returns
        -------
        ghi : np.ndarray
            2D array with shape (time, sites) in UTC.
        """
        if self._ghi is None:
            logger.debug('Calculating GHI.')
            self._ghi = (self.get_nsrdb_data('clearsky_ghi')
                         * self.clearsky_ratio)
            self._ghi[:, self.out_of_bounds] = 0
        return self._ghi

    @property
    def dni(self):
        """Get the dni (W/m2) which is clearsky dni (from NSRDB) when the GAN
        output is clear (clearsky_ratio is > cloud_threshold), and calculated
        from the DISC model when cloudy

        Returns
        -------
        dni : np.ndarray
            2D array with shape (time, sites) in UTC.
        """
        if self._dni is None:
            logger.debug('Calculating DNI.')
            self._dni = self.get_nsrdb_data('clearsky_dni')
            pressure = self.get_nsrdb_data('surface_pressure')
            doy = self.time_index.day_of_year.values
            cloudy_dni = disc(self.ghi, self.solar_zenith_angle, doy,
                              pressure=pressure)
            cloudy_dni = np.minimum(self._dni, cloudy_dni)
            self._dni[self.cloud_mask] = cloudy_dni[self.cloud_mask]
            self._dni = dark_night(self._dni, self.solar_zenith_angle)
            self._dni[:, self.out_of_bounds] = 0
        return self._dni

    @property
    def dhi(self):
        """Get the dhi (W/m2) which is calculated based on the simple
        relationship between GHI, DNI and solar zenith angle.

        Returns
        -------
        dhi : np.ndarray
            2D array with shape (time, sites) in UTC.
        """
        if self._dhi is None:
            logger.debug('Calculating DHI.')
            self._dhi, self._dni = calc_dhi(self.dni, self.ghi,
                                            self.solar_zenith_angle)
            self._dhi = dark_night(self._dhi, self.solar_zenith_angle)
            self._dhi[:, self.out_of_bounds] = 0
        return self._dhi

    @property
    def cloud_mask(self):
        """Get the cloud mask (True if cloudy) based on the GAN output clearsky
        ratio in UTC.

        Returns
        -------
        cloud_mask : np.ndarray
            2D array with shape (time, sites) in UTC.
        """
        return self.clearsky_ratio < self.cloud_threshold

    def get_nsrdb_data(self, dset):
        """Get an NSRDB dataset with spatial index corresponding to the sup3r
        GAN output data, averaged across the agg_factor.

        Parameters
        ----------
        dset : str
            Name of dataset to retrieve from NSRDB source file

        Returns
        -------
        out : np.ndarray
            Dataset of shape (time, sites) where time and sites correspond to
            the same shape as the sup3r GAN output data and if agg_factor > 1
            the sites is an average across multiple NSRDB sites.
        """

        logger.debug('Retrieving "{}" from NSRDB source data.'.format(dset))
        out = None

        for idx in range(self.idnn.shape[1]):
            temp = self.nsrdb[dset, self.nsrdb_tslice, self.idnn[:, idx]]
            temp = temp.astype(np.float32)

            if out is None:
                out = temp
            else:
                out += temp

        out /= self.idnn.shape[1]

        return out

    @staticmethod
    def get_sup3r_fps(fp_pattern, ignore=None):
        """Get a list of file chunks to run in parallel based on a file pattern

        NOTE: it's assumed that all source files have the pattern
        sup3r_file_TTTTTT_SSSSSS.h5 where TTTTTT is the zero-padded temporal
        chunk index and SSSSSS is the zero-padded spatial chunk index.

        Parameters
        ----------
        fp_pattern : str
            Unix-style file*pattern that matches a set of spatiotemporally
            chunked sup3r forward pass output files.
        ignore : str | None
            Ignore all files that have this string in their filenames.

        Returns
        -------
        fp_sets : list
            List of file sets where each file set is 3 temporally sequential
            files over the same spatial chunk. Each file set overlaps with its
            neighbor such that fp_sets[0][-1] == fp_sets[1][0] (this is so
            Solar can do temporal padding when rolling from GAN local time
            output to UTC).
        t_slices : list
            List of t_slice arguments corresponding to fp_sets to pass to
            Solar class initialization that will slice and reduce the
            overlapping time axis when Solar outputs irradiance data.
        temporal_ids : list
            List of temporal id strings TTTTTT corresponding to the fp_sets
        spatial_ids : list
            List of spatial id strings SSSSSS corresponding to the fp_sets
        target_fps : list
            List of actual target files corresponding to fp_sets, so for
            example the file set fp_sets[10] sliced by t_slices[10] is designed
            to process target_fps[10]
        """

        all_fps = [fp for fp in glob.glob(fp_pattern) if fp.endswith('.h5')]
        if ignore is not None:
            all_fps = [fp for fp in all_fps
                       if ignore not in os.path.basename(fp)]

        all_fps = sorted(all_fps)

        source_dir = os.path.dirname(all_fps[0])
        source_fn_base = os.path.basename(all_fps[0]).replace('.h5', '')
        source_fn_base = '_'.join(source_fn_base.split('_')[:-2])

        all_id_spatial = [fp.replace('.h5', '').split('_')[-1]
                          for fp in all_fps]
        all_id_temporal = [fp.replace('.h5', '').split('_')[-2]
                           for fp in all_fps]

        all_id_spatial = sorted(list(set(all_id_spatial)))
        all_id_temporal = sorted(list(set(all_id_temporal)))

        fp_sets = []
        t_slices = []
        temporal_ids = []
        spatial_ids = []
        target_fps = []
        for idt, id_temporal in enumerate(all_id_temporal):
            start = 0
            single_chunk_id_temps = [id_temporal]

            if idt > 0:
                start = 24
                single_chunk_id_temps.insert(0, all_id_temporal[idt - 1])

            if idt < (len(all_id_temporal) - 1):
                single_chunk_id_temps.append(all_id_temporal[idt + 1])

            for id_spatial in all_id_spatial:
                single_fp_set = []
                for t_str in single_chunk_id_temps:
                    fp = os.path.join(source_dir, source_fn_base)
                    fp += f'_{t_str}_{id_spatial}.h5'
                    single_fp_set.append(fp)

                fp_target = os.path.join(source_dir, source_fn_base)
                fp_target += f'_{id_temporal}_{id_spatial}.h5'

                fp_sets.append(single_fp_set)
                t_slices.append(slice(start, start + 24))
                temporal_ids.append(id_temporal)
                spatial_ids.append(id_spatial)
                target_fps.append(fp_target)

        return fp_sets, t_slices, temporal_ids, spatial_ids, target_fps

    @classmethod
    def get_node_cmd(cls, config):
        """Get a CLI call to run Solar.run_temporal_chunk() on a single
        node based on an input config.

        Parameters
        ----------
        config : dict
            sup3r solar config with all necessary args and kwargs to
            run Solar.run_temporal_chunk() on a single node.
        """
        import_str = 'import time;\n'
        import_str += 'from reV.pipeline.status import Status;\n'
        import_str += 'from rex import init_logger;\n'
        import_str += f'from sup3r.solar import {cls.__name__};\n'

        fun_str = get_fun_call_str(cls.run_temporal_chunk, config)

        log_file = config.get('log_file', None)
        log_level = config.get('log_level', 'INFO')
        log_arg_str = (f'"sup3r", log_level="{log_level}"')
        if log_file is not None:
            log_arg_str += f', log_file="{log_file}"'

        cmd = (f"python -c \'{import_str}\n"
               "t0 = time.time();\n"
               f"logger = init_logger({log_arg_str});\n"
               f"{fun_str};\n"
               "t_elap = time.time() - t0;\n")

        job_name = config.get('job_name', None)
        if job_name is not None:
            status_dir = config.get('status_dir', None)
            status_file_arg_str = f'"{status_dir}", '
            status_file_arg_str += f'module="{ModuleName.SOLAR}", '
            status_file_arg_str += f'job_name="{job_name}", '
            status_file_arg_str += 'attrs=job_attrs'

            cmd += ('job_attrs = {};\n'.format(json.dumps(config)
                                               .replace("null", "None")
                                               .replace("false", "False")
                                               .replace("true", "True")))
            cmd += 'job_attrs.update({"job_status": "successful"});\n'
            cmd += 'job_attrs.update({"time": t_elap});\n'
            cmd += f'Status.make_job_file({status_file_arg_str})'

        cmd += (";\'\n")

        return cmd.replace('\\', '/')

    def write(self, fp_out, features=('ghi', 'dni', 'dhi')):
        """Write irradiance datasets (ghi, dni, dhi) to output h5 file.

        Parameters
        ----------
        fp_out : str
            Filepath to an output h5 file to write irradiance variables to.
            Parent directory will be created if it does not exist.
        features : list | tuple
            List of features to write to disk. These have to be attributes of
            the Solar class (ghi, dni, dhi).
        """

        if not os.path.exists(os.path.dirname(fp_out)):
            os.makedirs(os.path.dirname(fp_out), exist_ok=True)

        with RexOutputs(fp_out, 'w') as fh:
            fh.meta = self.gan_data.meta
            fh.time_index = self.time_index

            for feature in features:
                attrs = H5_ATTRS[feature]
                arr = getattr(self, feature, None)
                if arr is None:
                    msg = ('Feature "{}" was not available from Solar '
                           'module class.'.format(feature))
                    logger.error(msg)
                    raise AttributeError(msg)

                fh.add_dataset(fp_out, feature, arr,
                               dtype=attrs['dtype'],
                               attrs=attrs,
                               chunks=attrs['chunks'])
                logger.info(f'Added "{feature}" to output file.')

            run_attrs = self.gan_data.h5[self._sup3r_fps[0]].global_attrs
            run_attrs['nsrdb_source'] = self._nsrdb_fp
            fh.run_attrs = run_attrs

        logger.info(f'Finished writing file: {fp_out}')

    @classmethod
    def run_temporal_chunk(cls, fp_pattern, nsrdb_fp,
                           fp_out_suffix='irradiance', tz=-6, agg_factor=1,
                           nn_threshold=0.5, cloud_threshold=0.99,
                           features=('ghi', 'dni', 'dhi'),
                           temporal_id=None):
        """Run the solar module on all spatial chunks for a single temporal
        chunk corresponding to the fp_pattern. This typically gets run from the
        CLI.

        Parameters
        ----------
        fp_pattern : str
            Unix-style file*pattern that matches a set of spatiotemporally
            chunked sup3r forward pass output files.
        nsrdb_fp : str
            Filepath to NSRDB .h5 file containing clearsky_ghi, clearsky_dni,
            clearsky_dhi data.
        fp_out_suffix : str
            Suffix to add to the input sup3r source files when writing the
            processed solar irradiance data to new data files.
        t_slice : slice
            Slicing argument to slice the temporal axis of the sup3r_fps source
            data after doing the tz roll to UTC but before returning the
            irradiance variables. This can be used to effectively pad the solar
            irradiance calculation in UTC time. For example, if sup3r_fps is 3
            files each with 24 hours of data, t_slice can be slice(24, 48) to
            only output the middle day of irradiance data, but padded by the
            other two days for the UTC output.
        tz : int
            The timezone offset for the data in sup3r_fps. It is assumed that
            the GAN is trained on data in local time and therefore the output
            in sup3r_fps should be treated as local time. For example, -6 is
            CST which is default for CONUS training data.
        agg_factor : int
            Spatial aggregation factor for nsrdb-to-GAN-meta e.g. the number of
            NSRDB spatial pixels to average for a single sup3r GAN output site.
        nn_threshold : float
            The KDTree nearest neighbor threshold that determines how far the
            sup3r GAN output data has to be from the NSRDB source data to get
            irradiance=0. Note that is value is in decimal degrees which is a
            very approximate way to determine real distance.
        cloud_threshold : float
            Clearsky ratio threshold below which the data is considered cloudy
            and DNI is calculated using DISC.
        features : list | tuple
            List of features to write to disk. These have to be attributes of
            the Solar class (ghi, dni, dhi).
        temporal_id : str | None
            One of the unique zero-padded temporal id's from the file chunks
            that match fp_pattern. This input typically gets set from the CLI.
            If None, this will run all temporal indices.
        """

        temp = cls.get_sup3r_fps(fp_pattern, ignore=f'_{fp_out_suffix}.h5')
        fp_sets, t_slices, temporal_ids, _, target_fps = temp

        if temporal_id is not None:
            fp_sets = [fp_set for i, fp_set in enumerate(fp_sets)
                       if temporal_ids[i] == temporal_id]
            t_slices = [t_slice for i, t_slice in enumerate(t_slices)
                        if temporal_ids[i] == temporal_id]
            target_fps = [target_fp for i, target_fp in enumerate(target_fps)
                          if temporal_ids[i] == temporal_id]

        zip_iter = zip(fp_sets, t_slices, target_fps)
        for i, (fp_set, t_slice, fp_target) in enumerate(zip_iter):
            fp_out = fp_target.replace('.h5', f'_{fp_out_suffix}.h5')
            logger.info('Running temporal index {} out of {}.'
                        .format(i + 1, len(fp_sets)))
            kwargs = dict(t_slice=t_slice,
                          tz=tz,
                          agg_factor=agg_factor,
                          nn_threshold=nn_threshold,
                          cloud_threshold=cloud_threshold)
            with Solar(fp_set, nsrdb_fp, **kwargs) as solar:
                solar.write(fp_out, features=features)



"""Sup3r topography utilities"""

import numpy as np
import logging
from scipy.spatial import KDTree
from rex import Resource
from abc import ABC, abstractmethod

import sup3r.preprocessing.data_handling
from sup3r.preprocessing.data_handling import DataHandlerNC, DataHandlerH5
from sup3r.postprocessing.file_handling import OutputHandler
from sup3r.utilities.utilities import get_source_type


logger = logging.getLogger(__name__)


class TopoExtract(ABC):
    """Class to extract high-res (4km+) topography rasters for new
    spatially-enhanced datasets (e.g. GCM files after spatial enhancement)
    using nearest neighbor mapping and aggregation from NREL datasets
    (e.g. WTK or NSRDB)
    """

    def __init__(self, file_paths, topo_source, s_enhance, agg_factor,
                 target=None, shape=None, raster_file=None, max_delta=20,
                 input_handler=None, ti_workers=1):
        """
        Parameters
        ----------
        file_paths : str | list
            A single source h5 file to extract raster data from or a list
            of netcdf files with identical grid. The string can be a unix-style
            file path which will be passed through glob.glob. This is
            typically low-res WRF output or GCM netcdf data files that is
            source low-resolution data intended to be sup3r resolved.
        topo_source : str
            Filepath to source wtk or nsrdb file to get hi-res (2km or 4km)
            elevation data from which will be mapped to the enhanced grid of
            the file_paths input
        s_enhance : int
            Factor by which the Sup3rGan model will enhance the spatial
            dimensions of low resolution data from file_paths input. For
            example, if file_paths has 100km data and s_enhance is 4, this
            class will output a topography raster corresponding to the
            file_paths grid enhanced 4x to ~25km
        agg_factor : int
            Factor by which to aggregate the topo_source_h5 elevation
            data to the resolution of the file_paths input enhanced by
            s_enhance. For example, if file_paths has 100km data and s_enhance
            is 4 resulting in a desired resolution of ~25km and topo_source_h5
            has a resolution of 4km, the agg_factor should be 36 so that 6x6
            4km cells are averaged to the ~25km enhanced grid.
        target : tuple
            (lat, lon) lower left corner of raster. Either need target+shape or
            raster_file.
        shape : tuple
            (rows, cols) grid size. Either need target+shape or raster_file.
        raster_file : str | None
            File for raster_index array for the corresponding target and shape.
            If specified the raster_index will be loaded from the file if it
            exists or written to the file if it does not yet exist.  If None
            raster_index will be calculated directly. Either need target+shape
            or raster_file.
        input_handler : str
            data handler class to use for input data. Provide a string name to
            match a class in data_handling.py. If None the correct handler will
            be guessed based on file type and time series properties.
        ti_workers : int | None
            max number of workers to use to get full time index. Useful when
            there are many input files each with a single time step. If this is
            greater than one, time indices for input files will be extracted in
            parallel and then concatenated to get the full time index. If input
            files do not all have time indices or if there are few input files
            this should be set to one.

        """

        logger.info('Initializing TopoExtract utility.')

        self._topo_source = topo_source
        self._s_enhance = s_enhance
        self._agg_factor = agg_factor
        self._tree = None
        self.ti_workers = ti_workers

        if input_handler is None:
            in_type = get_source_type(file_paths)
            if in_type == 'nc':
                input_handler = DataHandlerNC
            elif in_type == 'h5':
                input_handler = DataHandlerH5
            else:
                msg = ('Did not recognize input type "{}" for file paths: {}'
                       .format(in_type, file_paths))
                logger.error(msg)
                raise RuntimeError(msg)
        elif isinstance(input_handler, str):
            input_handler = getattr(sup3r.preprocessing.data_handling,
                                    input_handler, None)
            if input_handler is None:
                msg = ('Could not find requested data handler class '
                       f'"{input_handler}" in '
                       'sup3r.preprocessing.data_handling.')
                logger.error(msg)
                raise KeyError(msg)

        self.input_handler = input_handler(file_paths, [],
                                           target=target,
                                           shape=shape,
                                           raster_file=raster_file,
                                           max_delta=max_delta,
                                           ti_workers=ti_workers)

    @property
    @abstractmethod
    def source_elevation(self):
        """Get the 1D array of elevation data from the topo_source_h5"""

    @property
    @abstractmethod
    def source_lat_lon(self):
        """Get the 2D array (n, 2) of lat, lon data from the topo_source_h5"""

    @property
    def lr_shape(self):
        """Get the low-resolution spatial shape tuple"""
        return (self.lr_lat_lon.shape[0], self.lr_lat_lon.shape[1])

    @property
    def hr_shape(self):
        """Get the high-resolution spatial shape tuple"""
        return (self._s_enhance * self.lr_lat_lon.shape[0],
                self._s_enhance * self.lr_lat_lon.shape[1])

    @property
    def lr_lat_lon(self):
        """lat lon grid for data in format (spatial_1, spatial_2, 2) Lat/Lon
        array with same ordering in last dimension. This corresponds to the raw
        low-resolution meta data from the file_paths input.

        Returns
        -------
        ndarray
        """
        return self.input_handler.lat_lon

    @property
    def hr_lat_lon(self):
        """lat lon grid for data in format (spatial_1, spatial_2, 2) Lat/Lon
        array with same ordering in last dimension. This corresponds to the
        enhanced high-res meta data from the file_paths input * s_enhance.

        Returns
        -------
        ndarray
        """
        if self._s_enhance > 1:
            return OutputHandler.get_lat_lon(self.lr_lat_lon, self.hr_shape)
        else:
            return self.lr_lat_lon

    @property
    def tree(self):
        """Get the KDTree built on the source lat lon data"""
        if self._tree is None:
            self._tree = KDTree(self.source_lat_lon)
        return self._tree

    @property
    def nn(self):
        """Get the nearest neighbor indices """
        ll2 = np.vstack((self.hr_lat_lon[:, :, 0].flatten(),
                         self.hr_lat_lon[:, :, 1].flatten())).T
        _, nn = self.tree.query(ll2, k=self._agg_factor)
        if len(nn.shape) == 1:
            nn = np.expand_dims(nn, 1)
        return nn

    @property
    def hr_elev(self):
        """Get a raster of elevation values corresponding to the
        high-resolution grid (the file_paths input grid * s_enhance). The shape
        is (rows, cols)
        """
        nn = self.nn
        hr_elev = []
        for j in range(self._agg_factor):
            elev = self.source_elevation[nn[:, j]]
            elev = elev.reshape(self.hr_shape)
            hr_elev.append(elev)
        hr_elev = np.dstack(hr_elev).mean(axis=-1)
        logger.info('Finished mapping topo raster from {}'
                    .format(self._topo_source))
        return hr_elev

    @classmethod
    def get_topo_raster(cls, file_paths, topo_source, s_enhance,
                        agg_factor, target=None, shape=None, raster_file=None,
                        max_delta=20, input_handler=None):
        """Get the topography raster corresponding to the spatially enhanced
        grid from the file_paths input

        Parameters
        ----------
        file_paths : str | list
            A single source h5 wind file to extract raster data from or a list
            of netcdf files with identical grid. The string can be a unix-style
            file path which will be passed through glob.glob
        topo_source : str
            Filepath to source wtk, nsrdb, or netcdf file to get hi-res (2km or
            4km) data from which will be mapped to the enhanced grid of the
            file_paths input
        s_enhance : int
            Factor by which the Sup3rGan model will enhance the spatial
            dimensions of low resolution data from file_paths input. For
            example, if file_paths has 100km data and s_enhance is 4, this
            class will output a topography raster corresponding to the
            file_paths grid enhanced 4x to ~25km
        agg_factor : int
            Factor by which to aggregate the topo_source_h5 elevation data to
            the resolution of the file_paths input enhanced by s_enhance. For
            example, if file_paths has 100km data and s_enhance is 4 resulting
            in a desired resolution of ~25km and topo_source_h5 has a
            resolution of 4km, the agg_factor should be 6 so that 6x 4km cells
            are averaged to the ~25km enhanced grid.
        target : tuple
            (lat, lon) lower left corner of raster. Either need target+shape or
            raster_file.
        shape : tuple
            (rows, cols) grid size. Either need target+shape or raster_file.
        raster_file : str | None
            File for raster_index array for the corresponding target and shape.
            If specified the raster_index will be loaded from the file if it
            exists or written to the file if it does not yet exist.  If None
            raster_index will be calculated directly. Either need target+shape
            or raster_file.
        input_handler : str
            data handler class to use for input data. Provide a string name to
            match a class in data_handling.py. If None the correct handler will
            be guessed based on file type and time series properties.

        Returns
        -------
        topo_raster : np.ndarray
            Elevation raster with shape (hr_rows, hr_cols) corresponding to the
            shape of the spatially enhanced grid from file_paths * s_enhance.
            The elevation units correspond to the source units in
            topo_source_h5, usually meters.
        """

        te = cls(file_paths, topo_source, s_enhance, agg_factor,
                 target=target, shape=shape, raster_file=raster_file,
                 max_delta=max_delta, input_handler=input_handler)

        return te.hr_elev


class TopoExtractH5(TopoExtract):
    """TopoExtract for H5 files"""

    @property
    def source_elevation(self):
        """Get the 1D array of elevation data from the topo_source_h5"""
        with Resource(self._topo_source) as res:
            elev = res.get_meta_arr('elevation')
        return elev

    @property
    def source_lat_lon(self):
        """Get the 2D array (n, 2) of lat, lon data from the topo_source_h5"""
        with Resource(self._topo_source) as res:
            source_lat_lon = res.lat_lon
        return source_lat_lon


class TopoExtractNC(TopoExtract):
    """TopoExtract for netCDF files"""

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        args : list
            Same positional arguments as TopoExtract
        kwargs : dict
            Same keyword arguments as TopoExtract
        """

        super().__init__(*args, **kwargs)
        self.source_handler = DataHandlerNC(self._topo_source,
                                            features=['topography'],
                                            ti_workers=self.ti_workers)

    @property
    def source_elevation(self):
        """Get the 1D array of elevation data from the topo_source_h5"""
        elev = self.source_handler.data.reshape((-1))
        return elev

    @property
    def source_lat_lon(self):
        """Get the 2D array (n, 2) of lat, lon data from the topo_source_h5"""
        source_lat_lon = self.source_handler.lat_lon.reshape((-1, 2))
        return source_lat_lon



# -*- coding: utf-8 -*-
"""Special models for surface meteorological data."""
import logging
from fnmatch import fnmatch
import numpy as np
from PIL import Image
from sklearn import linear_model
from warnings import warn

from sup3r.models.abstract import AbstractSup3rGan
from sup3r.utilities.utilities import spatial_coarsening

logger = logging.getLogger(__name__)


class SurfaceSpatialMetModel(AbstractSup3rGan):
    """Model to spatially downscale daily-average near-surface temperature,
    relative humidity, and pressure

    Note that this model can also operate on temperature_*m,
    relativehumidity_*m, and pressure_*m datasets that are not
    strictly at the surface but could be near hub height.
    """

    TEMP_LAPSE = 6.5 / 1000
    """Temperature lapse rate: change in degrees C/K per meter"""

    PRES_DIV = 44307.69231
    """Air pressure scaling equation divisor variable:
    101325 * (1 - (1 - topo / PRES_DIV)**PRES_EXP)"""

    PRES_EXP = 5.25328
    """Air pressure scaling equation exponent variable:
    101325 * (1 - (1 - topo / PRES_DIV)**PRES_EXP)"""

    W_DELTA_TEMP = -3.99242830
    """Weight for the delta-temperature feature for the relative humidity
    linear regression model."""

    W_DELTA_TOPO = -0.01736911
    """Weight for the delta-topography feature for the relative humidity linear
    regression model."""

    def __init__(self, features, s_enhance, noise_adders=None,
                 temp_lapse=None, w_delta_temp=None, w_delta_topo=None,
                 pres_div=None, pres_exp=None):
        """
        Parameters
        ----------
        features : list
            List of feature names that this model will operate on for both
            input and output. This must match the feature axis ordering in the
            array input to generate(). Typically this is a list containing:
            temperature_*m, relativehumidity_*m, and pressure_*m. The list can
            contain multiple instances of each variable at different heights.
            relativehumidity_*m entries must have corresponding temperature_*m
            entires at the same hub height.
        s_enhance : int
            Integer factor by which the spatial axes are to be enhanced.
        noise_adders : float | list | None
            Option to add gaussian noise to spatial model output. Noise will be
            normally distributed with mean of 0 and standard deviation =
            noise_adders. noise_adders can be a single value or a list
            corresponding to the features list. None is no noise. The addition
            of noise has been shown to help downstream temporal-only models
            produce diurnal cycles in regions where there is minimal change in
            topography. A noise_adders around 0.07C (temperature) and 0.1%
            (relative humidity) have been shown to be effective.
        temp_lapse : None | float
            Temperature lapse rate: change in degrees C/K per meter. Defaults
            to the cls.TEMP_LAPSE attribute.
        w_delta_temp : None | float
            Weight for the delta-temperature feature for the relative humidity
            linear regression model. Defaults to the cls.W_DELTA_TEMP
            attribute.
        w_delta_topo : None | float
            Weight for the delta-topography feature for the relative humidity
            linear regression model. Defaults to the cls.W_DELTA_TOPO
            attribute.
        pres_div : None | float
            Divisor factor in the pressure scale height equation. Defaults to
            the cls.PRES_DIV attribute.
        pres_div : None | float
            Exponential factor in the pressure scale height equation. Defaults
            to the cls.PRES_EXP attribute.
        """

        self._features = features
        self._s_enhance = s_enhance
        self._noise_adders = noise_adders
        self._temp_lapse = temp_lapse or self.TEMP_LAPSE
        self._w_delta_temp = w_delta_temp or self.W_DELTA_TEMP
        self._w_delta_topo = w_delta_topo or self.W_DELTA_TOPO
        self._pres_div = pres_div or self.PRES_DIV
        self._pres_exp = pres_exp or self.PRES_EXP

        if isinstance(self._noise_adders, (int, float)):
            self._noise_adders = [self._noise_adders] * len(self._features)

    def __len__(self):
        """Get number of model steps (match interface of MultiStepGan)"""
        return 1

    @classmethod
    def load(cls, features, s_enhance, verbose=False, **kwargs):
        """Load the GAN with its sub-networks from a previously saved-to output
        directory.

        Parameters
        ----------
        features : list
            List of feature names that this model will operate on for both
            input and output. This must match the feature axis ordering in the
            array input to generate(). Typically this is a list containing:
            temperature_*m, relativehumidity_*m, and pressure_*m. The list can
            contain multiple instances of each variable at different heights.
            relativehumidity_*m entries must have corresponding temperature_*m
            entires at the same hub height.
        s_enhance : int
            Integer factor by which the spatial axes are to be enhanced.
        verbose : bool
            Flag to log information about the loaded model.
        kwargs : None | dict
            Optional kwargs to initialize SurfaceSpatialMetModel

        Returns
        -------
        out : SurfaceSpatialMetModel
            Returns an initialized SurfaceSpatialMetModel
        """

        model = cls(features, s_enhance, **kwargs)

        if verbose:
            logger.info('Loading SurfaceSpatialMetModel with meta data: {}'
                        .format(model.meta))

        return model

    @staticmethod
    def _get_s_enhance(topo_lr, topo_hr):
        """Get the spatial enhancement factor given low-res and high-res
        spatial rasters.

        Parameters
        ----------
        topo_lr : np.ndarray
            low-resolution surface elevation data in meters with shape
            (lat, lon)
        topo_hr : np.ndarray
            High-resolution surface elevation data in meters with shape
            (lat, lon)

        Returns
        -------
        s_enhance : int
            Integer factor by which the spatial axes are to be enhanced.
        """
        assert len(topo_lr.shape) == 2, 'topo_lr must be 2D'
        assert len(topo_hr.shape) == 2, 'topo_hr must be 2D'

        se0 = topo_hr.shape[0] / topo_lr.shape[0]
        se1 = topo_hr.shape[1] / topo_lr.shape[1]

        assert se0 % 1 == 0, f'Bad calculated s_enhance on axis 0: {se0}'
        assert se1 % 1 == 0, f'Bad calculated s_enhance on axis 1: {se1}'
        assert se0 == se1, 'Calculated s_enhance does not match along axis'

        return int(se0)

    @property
    def feature_inds_temp(self):
        """Get the feature index values for the temperature features."""
        inds = [i for i, name in enumerate(self._features)
                if fnmatch(name, 'temperature_*')]
        return inds

    @property
    def feature_inds_pres(self):
        """Get the feature index values for the pressure features."""
        inds = [i for i, name in enumerate(self._features)
                if fnmatch(name, 'pressure_*')]
        return inds

    @property
    def feature_inds_rh(self):
        """Get the feature index values for the relative humidity features."""
        inds = [i for i, name in enumerate(self._features)
                if fnmatch(name, 'relativehumidity_*')]
        return inds

    def _get_temp_ind(self, idf_rh):
        """Get the feature index value for the temperature feature
        corresponding to a relative humidity feature at the same hub height.

        Parameters
        ----------
        idf_rh : int
            Index in the feature list for a relativehumidity_*m feature

        Returns
        -------
        idf_temp : int
            Index in the feature list for a temperature_*m feature with the
            same hub height as the idf_rh input.
        """
        name_rh = self._features[idf_rh]
        suffix = name_rh.split('_')[-1]
        idf_temp = None
        for i in self.feature_inds_temp:
            if self._features[i].endswith(suffix):
                idf_temp = i
                break

        if idf_temp is None:
            msg = ('Could not find temperature feature corresponding to '
                   '"{}" in feature list: {}'.format(name_rh, self._features))
            logger.error(msg)
            raise KeyError(msg)

        return idf_temp

    @staticmethod
    def downscale_arr(arr, s_enhance, method=Image.Resampling.BILINEAR):
        """Downscale a 2D array of data Image.resize() method

        Parameters
        ----------
        arr : np.ndarray
            2D raster data, typically spatial daily average data with shape
            (lat, lon)
        s_enhance : int
            Integer factor by which the spatial axes are to be enhanced.
        method : Image.Resampling.BILINEAR
            An Image.Resampling method (NEAREST, BILINEAR, BICUBIC, LANCZOS).
            BILINEAR is default and has been tested to work best for
            SurfaceSpatialMetModel.
        """
        im = Image.fromarray(arr)
        im = im.resize((arr.shape[1] * s_enhance, arr.shape[0] * s_enhance),
                       resample=method)
        out = np.array(im)
        return out

    def downscale_temp(self, single_lr_temp, topo_lr, topo_hr):
        """Downscale temperature raster data at a single observation.

        This model uses a simple lapse rate that adjusts temperature as a
        function of elevation. The process is as follows:
            - add a scale factor to the low-res temperature field:
              topo_lr * TEMP_LAPSE
            - perform bilinear interpolation of the scaled temperature field.
            - subtract the scale factor from the high-res temperature field:
              topo_hr * TEMP_LAPSE

        Parameters
        ----------
        single_lr_temp : np.ndarray
            Single timestep temperature (deg C) raster data with shape
            (lat, lon) matching the low-resolution input data.
        topo_lr : np.ndarray
            low-resolution surface elevation data in meters with shape
            (lat, lon)
        topo_hr : np.ndarray
            High-resolution surface elevation data in meters with shape
            (lat, lon)

        Returns
        -------
        hi_res_temp : np.ndarray
            Single timestep temperature (deg C) raster data with shape
            (lat, lon) matching the high-resolution output data.
        """

        assert len(single_lr_temp.shape) == 2, 'Bad shape for single_lr_temp'
        assert len(topo_lr.shape) == 2, 'Bad shape for topo_lr'
        assert len(topo_hr.shape) == 2, 'Bad shape for topo_hr'

        lower_data = single_lr_temp + topo_lr * self._temp_lapse
        hi_res_temp = self.downscale_arr(lower_data, self._s_enhance)
        hi_res_temp -= topo_hr * self._temp_lapse

        return hi_res_temp

    def downscale_rh(self, single_lr_rh, single_lr_temp, single_hr_temp,
                     topo_lr, topo_hr):
        """Downscale relative humidity raster data at a single observation.

        Here's a description of the humidity scaling model:
            - Take low-resolution and high-resolution daily average
              temperature, relative humidity, and topography data.
            - Downscale all low-resolution variables using a bilinear image
              enhancement
            - The target model output is the difference between the
              interpolated low-res relative humidity and the true high-res
              relative humidity. Calculate this difference as a linear function
              of the same difference in the temperature and topography fields.
            - Use this linear regression model to calculate high-res relative
              humidity fields when provided low-res input of relative humidity
              along with low-res/high-res pairs of temperature and topography.

        Parameters
        ----------
        single_lr_rh : np.ndarray
            Single timestep relative humidity (%) raster data with shape
            (lat, lon) matching the low-resolution input data.
        single_lr_temp : np.ndarray
            Single timestep temperature (deg C) raster data with shape
            (lat, lon) matching the low-resolution input data.
        single_hr_temp : np.ndarray
            Single timestep temperature (deg C) raster data with shape
            (lat, lon) matching the high-resolution output data.
        topo_lr : np.ndarray
            low-resolution surface elevation data in meters with shape
            (lat, lon)
        topo_hr : np.ndarray
            High-resolution surface elevation data in meters with shape
            (lat, lon)

        Returns
        -------
        hi_res_rh : np.ndarray
            Single timestep relative humidity (%) raster data with shape
            (lat, lon) matching the high-resolution output data.
        """

        assert len(single_lr_rh.shape) == 2, 'Bad shape for single_lr_rh'
        assert len(single_hr_temp.shape) == 2, 'Bad shape for single_hr_temp'
        assert len(topo_lr.shape) == 2, 'Bad shape for topo_lr'
        assert len(topo_hr.shape) == 2, 'Bad shape for topo_hr'

        interp_rh = self.downscale_arr(single_lr_rh, self._s_enhance)
        interp_temp = self.downscale_arr(single_lr_temp, self._s_enhance)
        interp_topo = self.downscale_arr(topo_lr, self._s_enhance)

        delta_temp = single_hr_temp - interp_temp
        delta_topo = topo_hr - interp_topo

        hi_res_rh = (interp_rh
                     + self._w_delta_temp * delta_temp
                     + self._w_delta_topo * delta_topo)

        return hi_res_rh

    def downscale_pres(self, single_lr_pres, topo_lr, topo_hr):
        """Downscale pressure raster data at a single observation.

        This model uses a simple exponential scale height that adjusts
        temperature as a function of elevation. The process is as follows:
            - add a scale factor to the low-res pressure field:
              101325 * (1 - (1 - topo_lr / PRES_DIV)**PRES_EXP)
            - perform bilinear interpolation of the scaled pressure field.
            - subtract the scale factor from the high-res pressure field:
              101325 * (1 - (1 - topo_hr / PRES_DIV)**PRES_EXP)

        Parameters
        ----------
        single_lr_pres : np.ndarray
            Single timestep pressure (Pa) raster data with shape
            (lat, lon) matching the low-resolution input data.
        topo_lr : np.ndarray
            low-resolution surface elevation data in meters with shape
            (lat, lon)
        topo_hr : np.ndarray
            High-resolution surface elevation data in meters with shape
            (lat, lon)

        Returns
        -------
        hi_res_pres : np.ndarray
            Single timestep pressure (Pa) raster data with shape
            (lat, lon) matching the high-resolution output data.
        """

        if np.max(single_lr_pres) < 10000:
            msg = ('Pressure data appears to not be in Pa with min/mean/max: '
                   '{:.1f}/{:.1f}/{:.1f}'
                   .format(single_lr_pres.min(), single_lr_pres.mean(),
                           single_lr_pres.max()))
            logger.warning(msg)
            warn(msg)

        const = 101325 * (1 - (1 - topo_lr / self._pres_div)**self._pres_exp)
        single_lr_pres += const

        if np.min(single_lr_pres) < 0.0:
            msg = ('Spatial interpolation of surface pressure '
                   'resulted in negative values. Incorrectly '
                   'scaled/unscaled values or incorrect units are '
                   'the most likely causes.')
            logger.error(msg)
            raise ValueError(msg)

        hi_res_pres = self.downscale_arr(single_lr_pres, self._s_enhance)

        const = 101325 * (1 - (1 - topo_hr / self._pres_div)**self._pres_exp)
        hi_res_pres -= const

        if np.min(hi_res_pres) < 0.0:
            msg = ('Spatial interpolation of surface pressure '
                   'resulted in negative values. Incorrectly '
                   'scaled/unscaled values or incorrect units are '
                   'the most likely causes.')
            logger.error(msg)
            raise ValueError(msg)

        return hi_res_pres

    # pylint: disable=unused-argument
    def generate(self, low_res, norm_in=False, un_norm_out=False,
                 exogenous_data=None):
        """Use the generator model to generate high res data from low res
        input. This is the public generate function.

        Parameters
        ----------
        low_res : np.ndarray
            Low-resolution spatial input data, a 4D array of shape:
            (n_obs, spatial_1, spatial_2, 2), Where the feature channel is:
            [temperature_2m, relativehumidity_2m]
        norm_in : bool
            This doesnt do anything for this SurfaceSpatialMetModel, but is
            kept to keep the same interface as Sup3rGan
        un_norm_out : bool
            This doesnt do anything for this SurfaceSpatialMetModel, but is
            kept to keep the same interface as Sup3rGan
        exogenous_data : list
            For the SurfaceSpatialMetModel, this must be a 2-entry list where
            the first entry is a 2D (lat, lon) array of low-resolution surface
            elevation data in meters (must match spatial_1, spatial_2 from
            low_res), and the second entry is a 2D (lat, lon) array of
            high-resolution surface elevation data in meters.

        Returns
        -------
        hi_res : ndarray
            high-resolution spatial output data, a 4D array of shape:
            (n_obs, spatial_1, spatial_2, 2), Where the feature channel is:
            [temperature_2m, relativehumidity_2m]
        """

        msg = ('exogenous_data is of a bad type {}!'
               .format(type(exogenous_data)))
        assert isinstance(exogenous_data, (list, tuple)), msg
        msg = ('exogenous_data is of a bad length {}!'
               .format(len(exogenous_data)))
        assert len(exogenous_data) == 2, msg

        topo_lr = exogenous_data[0]
        topo_hr = exogenous_data[1]
        logger.debug('SurfaceSpatialMetModel received low/high res topo '
                     'shapes of {} and {}'
                     .format(topo_lr.shape, topo_hr.shape))

        msg = ('topo_lr has a bad shape {} that doesnt match the low res '
               'data shape {}'.format(topo_lr.shape, topo_hr.shape))
        assert isinstance(topo_lr, np.ndarray), msg
        assert isinstance(topo_hr, np.ndarray), msg
        assert len(topo_lr.shape) == 2, msg
        assert len(topo_hr.shape) == 2, msg
        assert topo_lr.shape[0] == low_res.shape[1], msg
        assert topo_lr.shape[1] == low_res.shape[2], msg
        s_enhance = self._get_s_enhance(topo_lr, topo_hr)
        msg = ('Topo shapes of {} and {} did not match desired spatial '
               'enhancement of {}'
               .format(topo_lr.shape, topo_hr.shape, self._s_enhance))
        assert self._s_enhance == s_enhance, msg

        hr_shape = (len(low_res),
                    int(low_res.shape[1] * self._s_enhance),
                    int(low_res.shape[2] * self._s_enhance),
                    len(self.output_features))
        logger.debug('SurfaceSpatialMetModel with s_enhance of {} '
                     'downscaling low-res shape {} to high-res shape {}'
                     .format(self._s_enhance, low_res.shape, hr_shape))

        hi_res = np.zeros(hr_shape, dtype=np.float32)
        for iobs in range(len(low_res)):
            for idf_temp in self.feature_inds_temp:
                _tmp = self.downscale_temp(low_res[iobs, :, :, idf_temp],
                                           topo_lr, topo_hr)
                hi_res[iobs, :, :, idf_temp] = _tmp

            for idf_pres in self.feature_inds_pres:
                _tmp = self.downscale_pres(low_res[iobs, :, :, idf_pres],
                                           topo_lr, topo_hr)
                hi_res[iobs, :, :, idf_pres] = _tmp

            for idf_rh in self.feature_inds_rh:
                idf_temp = self._get_temp_ind(idf_rh)
                _tmp = self.downscale_rh(low_res[iobs, :, :, idf_rh],
                                         low_res[iobs, :, :, idf_temp],
                                         hi_res[iobs, :, :, idf_temp],
                                         topo_lr, topo_hr)
                hi_res[iobs, :, :, idf_rh] = _tmp

        if self._noise_adders is not None:
            for idf, stdev in enumerate(self._noise_adders):
                if stdev is not None:
                    noise = np.random.normal(0, stdev, hi_res.shape[:-1])
                    hi_res[..., idf] += noise

        return hi_res

    @property
    def meta(self):
        """Get meta data dictionary that defines the model params"""
        return {'temp_lapse_rate': self._temp_lapse,
                's_enhance': self._s_enhance,
                't_enhance': 1,
                'noise_adders': self._noise_adders,
                'weight_for_delta_temp': self._w_delta_temp,
                'weight_for_delta_topo': self._w_delta_topo,
                'pressure_divisor': self._pres_div,
                'pressure_exponent': self._pres_exp,
                'training_features': self.training_features,
                'output_features': self.output_features,
                'class': self.__class__.__name__,
                }

    @property
    def training_features(self):
        """Get the list of input feature names that the generative model was
        trained on.

        Note that topography needs to be passed into generate() as an exogenous
        data input.
        """
        return self._features

    @property
    def output_features(self):
        """Get the list of output feature names that the generative model
        outputs"""
        return self._features

    def train(self, true_hr_temp, true_hr_rh, true_hr_topo):
        """This method trains the relative humidity linear model. The
        temperature and surface lapse rate models are parameterizations taken
        from the NSRDB and are not trained.

        Parameters
        ----------
        true_hr_temp : np.ndarray
            True high-resolution daily average temperature data in a 3D array
            of shape (lat, lon, n_days)
        true_hr_rh : np.ndarray
            True high-resolution daily average relative humidity data in a 3D
            array of shape (lat, lon, n_days)
        true_hr_topo : np.ndarray
            High-resolution surface elevation data in meters with shape
            (lat, lon)

        Returns
        -------
        w_delta_temp : float
            Weight for the delta-temperature feature for the relative humidity
            linear regression model.
        w_delta_topo : float
            Weight for the delta-topography feature for the relative humidity
            linear regression model.
        """

        assert len(true_hr_temp.shape) == 3, 'Bad true_hr_temp shape'
        assert len(true_hr_rh.shape) == 3, 'Bad true_hr_rh shape'
        assert len(true_hr_topo.shape) == 2, 'Bad true_hr_topo shape'

        true_hr_topo = np.expand_dims(true_hr_topo, axis=-1)
        true_hr_topo = np.repeat(true_hr_topo, true_hr_temp.shape[-1], axis=-1)

        true_lr_temp = spatial_coarsening(true_hr_temp,
                                          s_enhance=self._s_enhance,
                                          obs_axis=False)
        true_lr_rh = spatial_coarsening(true_hr_rh,
                                        s_enhance=self._s_enhance,
                                        obs_axis=False)
        true_lr_topo = spatial_coarsening(true_hr_topo,
                                          s_enhance=self._s_enhance,
                                          obs_axis=False)

        interp_hr_temp = np.full(true_hr_temp.shape, np.nan, dtype=np.float32)
        interp_hr_rh = np.full(true_hr_rh.shape, np.nan, dtype=np.float32)
        interp_hr_topo = np.full(true_hr_topo.shape, np.nan, dtype=np.float32)

        for i in range(interp_hr_temp.shape[-1]):
            interp_hr_temp[..., i] = self.downscale_arr(true_lr_temp[..., i],
                                                        self._s_enhance)
            interp_hr_rh[..., i] = self.downscale_arr(true_lr_rh[..., i],
                                                      self._s_enhance)
            interp_hr_topo[..., i] = self.downscale_arr(true_lr_topo[..., i],
                                                        self._s_enhance)

        x1 = true_hr_temp - interp_hr_temp
        x2 = true_hr_topo - interp_hr_topo
        x = np.vstack((x1.flatten(), x2.flatten())).T
        y = (true_hr_rh - interp_hr_rh).flatten()

        regr = linear_model.LinearRegression()
        regr.fit(x, y)
        if np.abs(regr.intercept_) > 1e-6:
            msg = ('Relative humidity linear model should have an intercept '
                   'of zero but the model fit an intercept of {}'
                   .format(regr.intercept_))
            logger.warning(msg)
            warn(msg)

        w_delta_temp, w_delta_topo = regr.coef_[0], regr.coef_[1]

        return w_delta_temp, w_delta_topo



# -*- coding: utf-8 -*-
"""
Abstract class to define the required interface for Sup3r model subclasses
"""

from abc import ABC, abstractmethod
from phygnn import CustomNetwork


class AbstractSup3rGan(ABC):
    """
    Abstract class to define the required interface for Sup3r model subclasses

    Note that this only sets the required interfaces for a GAN that can be
    loaded from disk and used to predict synthetic outputs. The interface for
    models that can be trained will be set in another class.
    """

    @classmethod
    @abstractmethod
    def load(cls, model_dir, verbose=True):
        """Load the GAN with its sub-networks from a previously saved-to output
        directory.

        Parameters
        ----------
        model_dir
            Directory to load GAN model files from.
        verbose : bool
            Flag to log information about the loaded model.

        Returns
        -------
        out : BaseModel
            Returns a pretrained gan model that was previously saved to
            model_dir
        """

    @staticmethod
    def seed(s=0):
        """
        Set the random seed for reproducible results.

        Parameters
        ----------
        s : int
            Random seed
        """
        CustomNetwork.seed(s=s)

    @abstractmethod
    def generate(self, low_res):
        """Use the generator model to generate high res data from low res
        input. This is the public generate function.

        Parameters
        ----------
        low_res : np.ndarray
            Low-resolution input data, usually a 4D or 5D array of shape:
            (n_obs, spatial_1, spatial_2, n_features)
            (n_obs, spatial_1, spatial_2, n_temporal, n_features)

        Returns
        -------
        hi_res : ndarray
            Synthetically generated high-resolution data, usually a 4D or 5D
            array with shape:
            (n_obs, spatial_1, spatial_2, n_features)
            (n_obs, spatial_1, spatial_2, n_temporal, n_features)
        """

    @property
    def s_enhance(self):
        """Factor by which model will enhance spatial resolution. Used in
        model training during high res coarsening"""
        return self.meta.get('s_enhance', None)

    @property
    def t_enhance(self):
        """Factor by which model will enhance temporal resolution. Used in
        model training during high res coarsening"""
        return self.meta.get('t_enhance', None)

    @property
    @abstractmethod
    def meta(self):
        """Get meta data dictionary that defines how the model was created"""

    @property
    @abstractmethod
    def training_features(self):
        """Get the list of input feature names that the generative model was
        trained on."""

    @property
    @abstractmethod
    def output_features(self):
        """Get the list of output feature names that the generative model
        outputs and that the discriminator predicts on."""



# -*- coding: utf-8 -*-
"""Custom sup3r solar module. This primarily converts GAN output clearsky ratio
to GHI, DNI, and DHI using NSRDB data and utility modules like DISC"""

from .solar import Solar



# -*- coding: utf-8 -*-
"""
sup3r bias correction calculation CLI entry points.
"""
import copy
import click
import logging
import os

from reV.pipeline.status import Status

from rex.utilities.execution import SubprocessManager
from rex.utilities.hpc import SLURM
from rex.utilities.loggers import init_mult

import sup3r.bias.bias_calc
from sup3r.utilities import ModuleName
from sup3r.version import __version__
from sup3r.pipeline.config import BaseConfig


logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, verbose):
    """Sup3r bias calc Command Line Interface"""
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose


@main.command()
@click.option('--config_file', '-c', required=True,
              type=click.Path(exists=True),
              help='sup3r bias correction calculation config .json file.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config_file, verbose):
    """Run sup3r bias correction calculation from a config file."""
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose
    status_dir = os.path.dirname(os.path.abspath(config_file))
    ctx.obj['OUT_DIR'] = status_dir
    config = BaseConfig(config_file)
    config['status_dir'] = status_dir
    config_verbose = config.get('log_level', 'INFO')
    config_verbose = (config_verbose == 'DEBUG')
    verbose = any([verbose, config_verbose, ctx.obj['VERBOSE']])

    init_mult('sup3r_bias_calc', os.path.join(status_dir, 'logs/'),
              modules=[__name__, 'sup3r'], verbose=verbose)

    exec_kwargs = config.get('execution_control', {})
    log_pattern = config.get('log_pattern', None)
    if log_pattern is not None:
        os.makedirs(os.path.dirname(log_pattern), exist_ok=True)
        if '.log' not in log_pattern:
            log_pattern += '.log'
        if '{node_index}' not in log_pattern:
            log_pattern = log_pattern.replace('.log', '_{node_index}.log')

    hardware_option = exec_kwargs.pop('option', 'local')
    exec_kwargs['stdout_path'] = os.path.join(status_dir, 'stdout/')
    logger.debug('Found execution kwargs: {}'.format(exec_kwargs))
    logger.debug('Hardware run option: "{}"'.format(hardware_option))

    calc_class_name = config['bias_calc_class']
    BiasCalcClass = getattr(sup3r.bias.bias_calc, calc_class_name)

    jobs = config['jobs']
    for i_node, job in enumerate(jobs):
        node_config = copy.deepcopy(job)
        node_config['status_dir'] = status_dir
        node_config['log_file'] = (
            log_pattern if log_pattern is None
            else os.path.normpath(log_pattern.format(node_index=i_node)))
        name = ('sup3r_bias_{}_{}'.format(os.path.basename(status_dir),
                                          str(i_node).zfill(6)))
        ctx.obj['NAME'] = name
        node_config['job_name'] = name

        cmd = BiasCalcClass.get_node_cmd(node_config)

        cmd_log = '\n\t'.join(cmd.split('\n'))
        logger.debug(f'Running command:\n\t{cmd_log}')

        if hardware_option.lower() in ('eagle', 'slurm'):
            kickoff_slurm_job(ctx, cmd, **exec_kwargs)
        else:
            kickoff_local_job(ctx, cmd)


def kickoff_slurm_job(ctx, cmd, alloc='sup3r', memory=None, walltime=4,
                      feature=None, stdout_path='./stdout/'):
    """Run sup3r on HPC via SLURM job submission.

    Parameters
    ----------
    ctx : click.pass_context
        Click context object where ctx.obj is a dictionary
    cmd : str
        Command to be submitted in SLURM shell script. Example:
            'python -m sup3r.cli forward_pass -c <config_file>'
    alloc : str
        HPC project (allocation) handle. Example: 'sup3r'.
    memory : int
        Node memory request in GB.
    walltime : float
        Node walltime request in hours.
    feature : str
        Additional flags for SLURM job. Format is "--qos=high"
        or "--depend=[state:job_id]". Default is None.
    stdout_path : str
        Path to print .stdout and .stderr files.
    """

    name = ctx.obj['NAME']
    out_dir = ctx.obj['OUT_DIR']
    slurm_manager = ctx.obj.get('SLURM_MANAGER', None)
    if slurm_manager is None:
        slurm_manager = SLURM()
        ctx.obj['SLURM_MANAGER'] = slurm_manager

    status = Status.retrieve_job_status(out_dir,
                                        module=ModuleName.BIAS_CALC,
                                        job_name=name,
                                        hardware='slurm',
                                        subprocess_manager=slurm_manager)

    msg = 'sup3r bias calc CLI failed to submit jobs!'
    if status == 'successful':
        msg = ('Job "{}" is successful in status json found in "{}", '
               'not re-running.'.format(name, out_dir))
    elif 'fail' not in str(status).lower() and status is not None:
        msg = ('Job "{}" was found with status "{}", not resubmitting'
               .format(name, status))
    else:
        logger.info('Running sup3r bias calc on SLURM with node name "{}".'
                    .format(name))
        out = slurm_manager.sbatch(cmd,
                                   alloc=alloc,
                                   memory=memory,
                                   walltime=walltime,
                                   feature=feature,
                                   name=name,
                                   stdout_path=stdout_path)[0]
        if out:
            msg = ('Kicked off sup3r bias calc job "{}" (SLURM jobid #{}).'
                   .format(name, out))

        # add job to sup3r status file.
        Status.add_job(out_dir, module=ModuleName.BIAS_CALC,
                       job_name=name, replace=True,
                       job_attrs={'job_id': out, 'hardware': 'slurm'})

    click.echo(msg)
    logger.info(msg)


def kickoff_local_job(ctx, cmd):
    """Run sup3r bias calc locally.

    Parameters
    ----------
    ctx : click.pass_context
        Click context object where ctx.obj is a dictionary
    cmd : str
        Command to be submitted in shell script. Example:
            'python -m sup3r.cli forward_pass -c <config_file>'
    """

    name = ctx.obj['NAME']
    out_dir = ctx.obj['OUT_DIR']
    subprocess_manager = SubprocessManager
    status = Status.retrieve_job_status(out_dir,
                                        module=ModuleName.BIAS_CALC,
                                        job_name=name)
    msg = 'sup3r bias calc CLI failed to submit jobs!'
    if status == 'successful':
        msg = ('Job "{}" is successful in status json found in "{}", '
               'not re-running.'.format(name, out_dir))
    elif 'fail' not in str(status).lower() and status is not None:
        msg = ('Job "{}" was found with status "{}", not resubmitting'
               .format(name, status))
    else:
        logger.info('Running sup3r bias calc locally with job name "{}".'
                    .format(name))
        Status.add_job(out_dir, module=ModuleName.BIAS_CALC,
                       job_name=name, replace=True)
        subprocess_manager.submit(cmd)
        msg = ('Completed sup3r bias calc job "{}".'.format(name))

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running sup3r bias calc CLI')
        raise



# -*- coding: utf-8 -*-
"""Sup3r model software"""
import logging
import tensorflow as tf

from sup3r.models.base import Sup3rGan


logger = logging.getLogger(__name__)


class WindCC(Sup3rGan):
    """Wind climate change model.

    Modifications to standard Sup3rGan:
        - Hi res topography is expected as the last feature channel in the true
          data in the true batch observation. This topo channel is appended to
          the generated output so the discriminator can look at the wind fields
          compared to the associated hi res topo.
    """

    def set_model_params(self, **kwargs):
        """Set parameters used for training the model

        Parameters
        ----------
        kwargs : dict
            Keyword arguments including 'training_features', 'output_features',
            'smoothed_features', 's_enhance', 't_enhance', 'smoothing'
        """
        output_features = kwargs['output_features']
        msg = ('Last output feature from the data handler must be topography '
               'to train the WindCC model, but received output features: {}'
               .format(output_features))
        assert output_features[-1] == 'topography', msg
        output_features.remove('topography')
        kwargs['output_features'] = output_features
        super().set_model_params(**kwargs)

    @tf.function
    def calc_loss(self, hi_res_true, hi_res_gen, **kwargs):
        """Calculate the GAN loss function using generated and true high
        resolution data.

        Parameters
        ----------
        hi_res_true : tf.Tensor
            Ground truth high resolution spatiotemporal data.
        hi_res_gen : tf.Tensor
            Superresolved high resolution spatiotemporal data generated by the
            generative model.
        kwargs : dict
            Key word arguments for:
            Sup3rGan.calc_loss(hi_res_true, hi_res_gen, **kwargs)

        Returns
        -------
        loss : tf.Tensor
            0D tensor representing the loss value for the network being trained
            (either generator or one of the discriminators)
        loss_details : dict
            Namespace of the breakdown of loss components
        """

        # append the true topography to the generated synthetic wind data
        hi_res_gen = tf.concat((hi_res_gen, hi_res_true[..., -1:]), axis=-1)

        return super().calc_loss(hi_res_true, hi_res_gen, **kwargs)
